import socket
import os
import json
import shlex
import subprocess
import pickle
import struct


def send_msg(sock, msg):
    # Prefix each message with a 4-byte length (network byte order)
    msg = struct.pack(">I", len(msg)) + msg
    sock.sendall(msg)


def recv_msg(sock):
    # Read message length and unpack it into an integer
    raw_msglen = recvall(sock, 4)
    if not raw_msglen:
        return None
    msglen = struct.unpack(">I", raw_msglen)[0]
    # Read the message data
    return recvall(sock, msglen)


def recvall(sock, n):
    # Helper function to recv n bytes or return None if EOF is hit
    data = bytearray()
    while len(data) < n:
        packet = sock.recv(n - len(data))
        if not packet:
            return None
        data.extend(packet)
    return data


def encode_request(obj: dict):
    return pickle.dumps(obj)


def decode_request(b: bytes):
    return pickle.loads(b)


## TODO: SocketManager might need to handle errors more gracefully
class SocketManager:
    def __init__(self, server_address):
        ## Configure them outside
        self.buf_size = 8192

        # Make sure the socket does not already exist
        ## TODO: Is this necessary?
        try:
            os.unlink(server_address)
        except OSError:
            if os.path.exists(server_address):
                raise
        # log("SocketManager: Made sure that socket does not exist")

        # Create a UDS socket
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        # log("SocketManager: Created socket")

        self.sock.bind(server_address)
        # log("SocketManager: Successfully bound to socket")

        ## TODO: Check if we need to configure the back# log
        self.sock.listen()
        # log("SocketManager: Listenting on socket")

    def get_next_cmd(self):
        connection, client_address = self.sock.accept()
        data = connection.recv(self.buf_size)

        ## TODO: This could be avoided for efficiency
        str_data = data.decode("utf-8")
        # log("Received data:", str_data)
        ## TODO: Lift this requirement if needed
        ##
        ## We need to ensure that we read a command at once or the command was empty (only relevant in the first invocation)
        assert str_data.endswith("\n") or str_data == ""

        return str_data, connection

    ## This method respond to the connection we last got input from
    ## In the case of the UnixPipes, we don't have any state management here
    ##   since all reads/writes go to/from the same fifos
    def respond(self, message, connection):
        bytes_message = message.encode("utf-8")
        connection.sendall(bytes_message)
        connection.close()

    def close(self):
        self.sock.close()
        # log("SocketManager: Closed")



import socket

from ir_defs.arg import Arg
from util import log

## TODO: Resources should probably be more elaborate than just a
## string and a line range. They could be URLs, and possibly other things.


## TODO: Think if we can have any optimizations if we know the size of a resource.
class Resource:
    def __init__(self, uri):
        self.uri = uri

    def __repr__(self):
        output = str(self.uri)
        return output

    def is_stdin(self):
        return False

    def is_stdout(self):
        return False

    ## TODO: Make this check properly
    def __eq__(self, other):
        if isinstance(other, Resource):
            return self.uri == other.uri
        return False


class FileDescriptorResource(Resource):
    def __init__(self, fd):
        assert isinstance(fd, tuple) and len(fd) == 2 and fd[0] == "fd"
        self.uri = fd

    def is_stdin(self):
        return self.uri == ("fd", 0)

    def is_stdout(self):
        return self.uri == ("fd", 1)


class FileResource(Resource):
    ## The uri is the path of the file.
    def __init__(self, path):
        log("class of path", type(path))
        assert isinstance(path, Arg)
        ## TODO: Make sure that paths are normalized
        self.uri = path

    def __eq__(self, other):
        if isinstance(other, FileResource):
            return self.uri == other.uri
        return False


class TemporaryFileResource(Resource):
    def __init__(self):
        self.uri = None


# A FIFO.
class EphemeralResource(Resource):
    def __init__(self):
        self.uri = None


class RemoteFileResource(Resource):
    def __init__(self):
        raise NotImplementedError("RemoteFileResource is an interface")

    def __str__(self):
        raise NotImplementedError("RemoteFileResource is an interface")

    def is_available_on(self, host):
        raise NotImplementedError("RemoteFileResource is an interface")

    def _normalize_addr(self, addr):
        """
        Helper, removes port number if supplied in address and normalizes host address
        Example1: datanode1 -> 172.18.0.3
        Example2: 172.18.0.3:5555 -> 172.18.0.3
        """
        host = addr.rsplit(":", 1)[0]
        normalized_host = socket.gethostbyaddr(host)[2][0]
        return normalized_host


class HDFSFileResource(RemoteFileResource):
    ## The uri is the path of the file.
    def __init__(self, uri, resource_hosts):
        """
        Params:
            uri: Usually the path to the file. The path doesn't include the top directory
            which is different between hosts. The str function adds the prefix $HDFS_DATANODE_DIR/
            which should be defined on host machine worker environment.
            resource_hosts: the addresses of all the machines containing
            the resource.
        """
        self.uri = uri
        self.hosts = list(map(self._normalize_addr, resource_hosts))

    def __eq__(self, other):
        if isinstance(other, HDFSFileResource):
            return self.uri == other.uri
        return False

    def is_available_on(self, host):
        return host in self.hosts

    def __repr__(self):
        return f"hdfs://{self.uri}"

    def __str__(self):
        return f"$HDFS_DATANODE_DIR/{self.uri}"


# DFS logical split resource
class DFSSplitResource(RemoteFileResource):
    def __init__(self, config, config_path, split_num, hosts):
        self.config = config
        self.config_path = config_path
        self.split_num = split_num
        self.hosts = list(map(self._normalize_addr, hosts))

    def is_available_on(self, host):
        return host in self.hosts

    def set_config_path(self, config_path):
        self.config_path = config_path

    def __str__(self):
        return self.config_path



import os

import config
from ir_defs.dfg_node import DFGNode
from ir_defs.arg import Arg


class DFSSplitReader(DFGNode):
    def __init__(
        self,
        inputs,
        outputs,
        com_name,
        com_category,
        com_options=None,
        com_redirs=None,
        com_assignments=None,
    ):
        com_options = [] if com_options is None else com_options
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments

        super().__init__(
            inputs,
            outputs,
            com_name,
            com_category,
            com_options=com_options,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
        )

    def set_server_address(self, addr):  # ex addr: 127.0.0.1:50051
        self.com_options.append((3, Arg.string_to_arg(f"--addr {addr}")))


def make_dfs_split_reader_node(inputs, output, split_num, prefix):
    split_reader_bin = os.path.join(
        config.PASH_TOP, config.config["runtime"]["dfs_split_reader_binary"]
    )
    com_name = Arg.string_to_arg(split_reader_bin)
    com_category = "pure"
    options = []
    options.append((1, Arg.string_to_arg(f"--prefix '{prefix}'")))
    options.append((2, Arg.string_to_arg(f"--split {split_num}")))

    return DFSSplitReader(inputs, [output], com_name, com_category, options)



"""
Configuration module for PaSh compiler.

Constants (set at import time from environment or package location):
- PASH_TOP: Root directory of PaSh installation/package
- PASH_TMP_PREFIX: Temporary directory for this session
- BASH_VERSION: Tuple of bash version numbers
- HDFS_PREFIX: Prefix for HDFS paths

Runtime state (set during initialization):
- config: Dict loaded from config.json
- pash_args: Parsed command-line arguments
- LOGGING_PREFIX: Prefix for log messages
"""

import json
import logging
import os
import sys
from pathlib import Path


def get_package_root() -> Path:
    """Get the root directory of the pash package.

    This is the directory containing compiler/, preprocessor/, etc.
    When installed via pip, this is the site-packages/pash directory.
    """
    return Path(__file__).parent.parent


def get_pash_top() -> str:
    """Get PASH_TOP, preferring environment variable if set.

    This allows both pip-installed usage (uses package location)
    and development usage (uses PASH_TOP from pa.sh).
    """
    env_pash_top = os.getenv("PASH_TOP")
    if env_pash_top:
        return env_pash_top
    return str(get_package_root())


# === Constants from environment or package location ===

PASH_TOP = get_pash_top()

# PASH_TMP_PREFIX must be set by the entry point (pa.sh or cli.py)
_pash_tmp_prefix = os.getenv("PASH_TMP_PREFIX")
if _pash_tmp_prefix is None:
    # Fallback for when module is imported outside of normal execution
    import tempfile
    _pash_tmp_prefix = tempfile.mkdtemp(prefix="pash_") + "/"
PASH_TMP_PREFIX = _pash_tmp_prefix

# BASH_VERSION with fallback
_bash_version_str = os.getenv("PASH_BASH_VERSION", "5 0 0")
BASH_VERSION = tuple(int(i) for i in _bash_version_str.split(" "))

HDFS_PREFIX = "$HDFS_DATANODE_DIR/"

# === Runtime state ===

LOGGING_PREFIX = ""
config = {}
pash_args = None

# Increase recursion limit for parser/unparser
sys.setrecursionlimit(10000)


def set_config_globals_from_pash_args(given_pash_args):
    """Initialize runtime state from parsed arguments."""
    global pash_args
    pash_args = given_pash_args

    # Configure logging
    if given_pash_args.log_file == "":
        logging.basicConfig(format="%(message)s")
    else:
        logging.basicConfig(
            format="%(message)s",
            filename=os.path.abspath(given_pash_args.log_file),
            filemode="w",
        )

    # Set debug level
    if given_pash_args.debug == 0:
        logging.getLogger().setLevel(logging.ERROR)
    elif given_pash_args.debug == 1:
        logging.getLogger().setLevel(logging.WARNING)
    elif given_pash_args.debug == 2:
        logging.getLogger().setLevel(logging.INFO)
    elif given_pash_args.debug >= 3:
        logging.getLogger().setLevel(logging.DEBUG)


def load_config(config_file_path=""):
    """Load configuration from JSON file."""
    global config

    if config_file_path == "":
        # Look for config.json relative to this module (in compiler directory)
        config_file_path = str(Path(__file__).parent / "config.json")

    with open(config_file_path) as config_file:
        config = json.load(config_file)

    if not config:
        raise Exception(f"No valid configuration could be loaded from {config_file_path}")
    if "distr_planner" not in config:
        raise Exception(f"Missing `distr_planner` config in {config_file_path}")


def set_vars_file(var_file_path: str, var_dict: dict):
    """Set shell variables in config."""
    global config
    config["shell_variables"] = var_dict
    config["shell_variables_file_path"] = var_file_path



"""
Handler for ForNode that injects loop tracking code.

This is specific to the pash/compiler version which needs to
track loop iterations for speculative execution.
"""

from __future__ import annotations
import copy
from typing import TYPE_CHECKING

from shasta.ast_node import ForNode, make_typed_semi_sequence
from shasta.json_to_ast import to_ast_node
from shell_ast.ast_util import (
    make_export_var_constant_string,
    make_increment_var,
    export_pash_loop_iters_for_current_context,
)
from env_var_names import loop_iter_var

if TYPE_CHECKING:
    from shell_ast.walk_preprocess import PreprocessContext, NodeResult, WalkPreprocess


def for_node_with_loop_tracking(
    node: ForNode, ctx: "PreprocessContext", walker: "WalkPreprocess"
) -> "NodeResult":
    """
    Handler for ForNode that injects loop tracking code.

    This handler:
    1. Enters a new loop context
    2. Preprocesses the loop body
    3. Injects loop iteration counter initialization
    4. Injects loop iteration counter increment in the body
    5. Exports loop iteration context for the runtime
    6. Exits the loop context

    Args:
        node: The ForNode to process
        ctx: The preprocessing context
        walker: The walker instance for recursive calls

    Returns:
        NodeResult with the transformed node and metadata
    """
    # Import here to avoid circular dependency
    from shell_ast.walk_preprocess import NodeResult

    # Enter loop context
    loop_id = ctx.trans_options.enter_loop()

    # Preprocess the body using close-node semantics
    preprocessed_body, something_replaced = walker.walk_close(node.body, ctx)

    # Create loop tracking nodes
    var_name = loop_iter_var(loop_id)
    export_node = make_export_var_constant_string(var_name, "0")
    increment_node = make_increment_var(var_name)

    # Get all loop IDs for context export
    all_loop_ids = ctx.trans_options.get_current_loop_context()
    save_loop_iters_node = export_pash_loop_iters_for_current_context(all_loop_ids)

    # Modify the loop body to include tracking
    node.body = make_typed_semi_sequence(
        [
            to_ast_node(increment_node),
            to_ast_node(save_loop_iters_node),
            ## KK 2026-01-20 This deepcopy used to be there but led to serious
            ##               performance issues with some tests. 
            ##               I think it is not necessary but leaving it here in case we want to revert back.
            # copy.deepcopy(preprocessed_body),
            preprocessed_body,
        ]
    )

    # Exit loop context
    ctx.trans_options.exit_loop()

    # Reset loop iters after exiting
    out_of_loop_ids = ctx.trans_options.get_current_loop_context()
    reset_loop_iters_node = export_pash_loop_iters_for_current_context(out_of_loop_ids)

    # Wrap the entire for loop with initialization and reset
    new_node = make_typed_semi_sequence(
        [to_ast_node(export_node), node, to_ast_node(reset_loop_iters_node)]
    )

    return NodeResult(ast=new_node, something_replaced=something_replaced)



from pash_annotations.datatypes.AccessKind import make_stream_output, make_stream_input
from pash_annotations.datatypes.BasicDatatypes import ArgStringType
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

import config
import os

from annotations_utils.util_cmd_invocations import (
    to_arg_from_cmd_inv_with_io_vars_without_streaming_inputs_or_outputs_for_wrapping,
)
from ir_defs.dfg_node import DFGNode
from ir_defs.arg import Arg


class RWrap(DFGNode):
    def __init__(
        self,
        cmd_invocation_with_io_vars,
        com_redirs=None,
        com_assignments=None,
        parallelizer_list=None,
        cmd_related_properties=None,
        wrapped_node_name=None,
    ):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments
        self.wrapped_node_name = wrapped_node_name
        super().__init__(
            cmd_invocation_with_io_vars,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
            parallelizer_list=parallelizer_list,
            cmd_related_properties=cmd_related_properties,
        )

    ## Get the label of the node. By default, it is simply the name
    def get_dot_label(self) -> str:
        ## The name could be a full path
        name = self.cmd_invocation_with_io_vars.cmd_name
        basename = os.path.basename(str(name))

        wrapped_node_name = self.wrapped_node_name
        return f"{basename}({wrapped_node_name})"


def wrap_node(node: DFGNode, edges):
    r_wrap_bin = os.path.join(
        config.PASH_TOP, config.config["runtime"]["r_wrap_binary"]
    )

    ## At the moment we can only wrap a node that takes its input from stdin
    ## and outputs to stdout. Therefore the node needs to have only one input and one output.
    ## TO CHECK: with the remodelling also other cases should be handled
    inputs = node.get_input_list()
    assert len(inputs) == 1
    input_id = inputs[0]
    outputs = node.get_output_list()
    ## TODO: Would it make sense for outputs to be less than one?
    ## TODO: changed this from <= to == 1 to simplify reasoning later for now
    assert len(outputs) == 1
    output_id = outputs[0]
    access_map = {input_id: make_stream_input(), output_id: make_stream_output()}

    # create bash -c argument
    cmd_inv_with_io_vars: CommandInvocationWithIOVars = node.cmd_invocation_with_io_vars
    # do we need to copy here? currently, it seems fine
    cmd_inv_with_io_vars.remove_streaming_inputs()
    cmd_inv_with_io_vars.remove_streaming_outputs()
    # any non-streaming inputs or outputs are converted here already!
    cmd = to_arg_from_cmd_inv_with_io_vars_without_streaming_inputs_or_outputs_for_wrapping(
        cmd_inv_with_io_vars, edges
    )

    bash_command_arg = [Arg.string_to_arg("bash -c")]
    operand_list = bash_command_arg + [cmd]

    cmd_inv_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=r_wrap_bin,
        flag_option_list=[],
        operand_list=operand_list,
        implicit_use_of_streaming_input=input_id,
        implicit_use_of_streaming_output=output_id,
        access_map=access_map,
    )

    ## TODO: It is not clear if it is safe to just pass redirections and assignments down the line as is
    redirs = node.com_redirs
    assignments = node.com_assignments

    return RWrap(
        cmd_inv_with_io_vars,
        com_redirs=redirs,
        com_assignments=assignments,
        wrapped_node_name=node.cmd_invocation_with_io_vars.cmd_name,
    )



"""
Core compilation functionality for PaSh.

This module contains the compilation pipeline that transforms IR to optimized shell scripts.
It is used by compilation_server.py and can be imported directly for programmatic use.
"""

import pickle
import traceback
from datetime import datetime

from sh_expand import env_vars_util
from sh_expand.expand import ExpansionError

import config
from ir import IR, FileIdGen, make_cat_node
from ast_to_ir import compile_asts
from ir_to_ast import to_shell
from pash_graphviz import maybe_generate_graphviz
from util import log, print_time_delta, ptempfile, UnparallelizableError, NotAllRegionParallelizableError, AdjLineNotImplementedError

from ir_defs.nodes.eager import Eager
from ir_defs.nodes.hdfs_cat import HDFSCat
from ir_defs.nodes.pash_split import Split
from ir_defs.resource import DFSSplitResource

import ir_defs.nodes.r_split as r_split
import ir_defs.nodes.r_unwrap as r_unwrap
import ir_defs.nodes.dgsh_tee as dgsh_tee
import ir_defs.nodes.dfs_split_reader as dfs_split_reader

# Distributed Exec
import dspash.hdfs_utils as hdfs_utils

runtime_config = {}


class CompilerConfig:
    """Configuration for a single compilation run."""
    def __init__(self, width):
        self.width = width

    def __repr__(self):
        return f"CompilerConfig(Width:{self.width})"


def compile_ir(ir_filename, compiled_script_file, args, compiler_config):
    """
    Main compilation entry point.

    Args:
        ir_filename: Path to input IR file (pickle)
        compiled_script_file: Path to write compiled shell script
        args: Parsed CLI arguments
        compiler_config: Compilation configuration

    Returns:
        IR object on success, None otherwise.
    """
    ret = None
    try:
        ret = compile_optimize_output_script(
            ir_filename, compiled_script_file, args, compiler_config
        )
    except ExpansionError as e:
        log("WARNING: Exception caught because some region(s) are not expandable and therefore unparallelizable:", e)
        raise NotAllRegionParallelizableError()
    except UnparallelizableError as e:
        log("WARNING: Exception caught because some region(s) are unparallelizable:", e)
        raise NotAllRegionParallelizableError()
        # log(traceback.format_exc()) # uncomment for exact trace report (PaSh user should see informative messages for unparellizable regions)
    except (AdjLineNotImplementedError, NotImplementedError) as e:
        log("WARNING: Exception caught because some part is not implemented:", e)
        log(traceback.format_exc())
    except Exception as e:
        log("WARNING: Exception caught:", e)
        log(traceback.format_exc())

    return ret


def compile_optimize_output_script(
    ir_filename, compiled_script_file, args, compiler_config
):
    global runtime_config

    ret = None

    ## Load the df_region from a file
    candidate_df_region = load_df_region(ir_filename)

    ## Compile it
    optimized_ast_or_ir = compile_optimize_df_region(
        candidate_df_region, args, compiler_config
    )

    ## Call the backend that executes the optimized dataflow graph
    ## TODO: Should never be the case for now. This is obsolete.
    assert not runtime_config["distr_backend"]

    ## If the candidate DF region was indeed a DF region then we have an IR
    ## which should be translated to a parallel script.
    if isinstance(optimized_ast_or_ir, IR):
        if args.distributed_exec:
            ir_filename = ptempfile()
            script_to_execute = (
                f"$PASH_TOP/compiler/dspash/remote_exec_graph.sh {ir_filename}\n"
            )
            ## This might not be needed anymore (since the output script is output anyway)
            ## TODO: This is probably useless, remove
            maybe_log_optimized_script(script_to_execute, args)

            with open(ir_filename, "wb") as f:
                obj = (optimized_ast_or_ir, config.config["shell_variables"])
                pickle.dump(obj, f)
        else:
            script_to_execute = to_shell(optimized_ast_or_ir, args)

        log("Optimized script saved in:", compiled_script_file)
        with open(compiled_script_file, "w", encoding="utf-8") as f:
            f.write(script_to_execute)

        ret = optimized_ast_or_ir
    else:
        raise UnparallelizableError("Script failed to compile!")

    return ret


def load_df_region(ir_filename):
    log("Retrieving candidate DF region: {} ... ".format(ir_filename), end="")
    with open(ir_filename, "rb") as ir_file:
        candidate_df_region = pickle.load(ir_file)
    log("Done!")
    return candidate_df_region


def compile_optimize_df_region(df_region, args, compiler_config):
    """Compile and optimize a dataflow region."""
    ## Compile the candidate DF regions
    compilation_start_time = datetime.now()
    asts_and_irs = compile_candidate_df_region(df_region, config.config)
    compilation_end_time = datetime.now()
    print_time_delta("Compilation", compilation_start_time, compilation_end_time)

    ## Optimize all the IRs that can be optimized
    if args.no_optimize:
        optimized_asts_and_irs = asts_and_irs
    else:
        optimized_asts_and_irs = optimize_irs(asts_and_irs, args, compiler_config)

    ## TODO: Normally this could return more than one compiled ASTs (containing IRs in them).
    ##       To correctly handle that we would need to really replace the optimized IRs
    ##       with the final parallel corresponding scripts.
    ##
    ##       However, for now we just assume that there is one IR that we can execute as is.
    ##
    ## TODO: This might bite us with the quick-abort.
    ##       It might complicate things having a script whose half is compiled to a graph and its other half not.
    assert len(optimized_asts_and_irs) == 1
    optimized_ast_or_ir = optimized_asts_and_irs[0]

    return optimized_ast_or_ir


def maybe_log_optimized_script(script_to_execute, args):
    ## TODO: Merge this write with the one below. Maybe even move this logic in `pash_runtime.sh`
    ## Output the optimized shell script for inspection
    if args.output_optimized:
        output_script_path = runtime_config["optimized_script_filename"]
        with open(output_script_path, "w", encoding="utf-8") as output_script_file:
            log("Optimized script:")
            log(script_to_execute)
            output_script_file.write(script_to_execute)


def compile_candidate_df_region(candidate_df_region, config):
    ## This is for the files in the IR
    fileIdGen = FileIdGen()

    ## If the candidate DF region is not from the top level then
    ## it won't be a list and thus we need to make it into a list to compile it.
    if not isinstance(candidate_df_region, list):
        candidate_df_region = [candidate_df_region]

    ## Compile the asts
    ## TODO: Since compilation happens at runtime, we can now expand everything accordingly.
    ##       We can do that using a shell for start:
    ##         if a word is safe to expand, then call a shell to expand it.
    compiled_asts = compile_asts(candidate_df_region, fileIdGen, config)

    return compiled_asts


def optimize_irs(asts_and_irs, args, compiler_config):
    """Apply parallelization transformations to IRs."""
    global runtime_config

    optimization_start_time = datetime.now()

    optimized_asts_and_irs = []
    for ast_or_ir in asts_and_irs:
        if isinstance(ast_or_ir, IR):
            ## Assert that the graph that was returned from compilation is valid
            assert ast_or_ir.valid()

            # log(ir_node)
            # with cProfile.Profile() as pr:
            distributed_graph = choose_and_apply_parallelizing_transformations(
                ast_or_ir,
                compiler_config.width,
                runtime_config["batch_size"],
                args.r_split_batch_size,
            )
            # pr.print_stats()

            # Eagers are added in remote notes when using distributed exec
            if not args.no_eager and not args.distributed_exec:
                eager_distributed_graph = add_eager_nodes(distributed_graph)
            else:
                eager_distributed_graph = distributed_graph

            ## Assert that the graph stayed valid after all transformations
            assert eager_distributed_graph.valid()

            ## Print statistics of output nodes
            print_graph_statistics(eager_distributed_graph)

            optimized_asts_and_irs.append(eager_distributed_graph)
        else:
            optimized_asts_and_irs.append(ast_or_ir)

    optimization_end_time = datetime.now()
    print_time_delta("Optimization", optimization_start_time, optimization_end_time)

    return optimized_asts_and_irs


def print_graph_statistics(graph):
    total_nodes = graph.nodes
    eager_nodes = [node for node in total_nodes.values() if isinstance(node, Eager)]
    log("Total nodes after optimization:", len(total_nodes))
    log(" -- out of which:")
    log("Eager nodes:", len(eager_nodes))


def choose_and_apply_parallelizing_transformations(
    graph, fan_out, batch_size, r_split_batch_size
):
    parallelizer_map = choose_parallelizing_transformations(graph)
    apply_parallelizing_transformations(
        graph, parallelizer_map, fan_out, batch_size, r_split_batch_size
    )
    return graph


def choose_parallelizing_transformations(graph):  # shall return map
    source_node_ids = graph.source_nodes()
    parallelizer_map = {}
    workset = source_node_ids
    visited = set()
    # We apply a modified BFS such that we ensure that we know which parallelizer was chosen for all previous nodes
    # and assume that the decision for any subsequent node will exploit any potential synergy effects
    while len(workset) > 0:
        curr_id = workset.pop(0)
        assert isinstance(curr_id, int)
        all_previous_nodes_visited = all(
            prev in visited for prev in graph.get_previous_nodes(curr_id)
        )
        if not all_previous_nodes_visited:
            workset.append(curr_id)
        elif not curr_id in visited:
            next_node_ids = graph.get_next_nodes(curr_id)
            workset += next_node_ids
            parallelizer_map[curr_id] = choose_parallelizing_transformation(
                curr_id, graph
            )
            visited.add(curr_id)
    return parallelizer_map


## This currently chooses the best parallelization based on priority:
## 1. The round robin
## 2. The round robin after having performed unwrap (not sure why this is the second priority)
## 3. The consecutive chunks
##
## TODO: In the future, we could develop more complex strategies
def choose_parallelizing_transformation(curr_id, graph):  # shall return map entry
    curr = graph.get_node(curr_id)
    list_all_parallelizers_in_priority = [
        curr.get_option_implemented_round_robin_parallelizer(),
        curr.get_option_implemented_round_robin_with_unwrap_parallelizer(),
        curr.get_option_implemented_consecutive_chunks_parallelizer(),
    ]
    return next(
        (item for item in list_all_parallelizers_in_priority if item is not None), None
    )


def apply_parallelizing_transformations(
    graph, parallelizer_map, fan_out, batch_size, r_split_batch_size
):
    fileIdGen = graph.get_file_id_gen()
    node_id_non_none_parallelizer_list = [
        (node_id, parallelizer)
        for (node_id, parallelizer) in parallelizer_map.items()
        if parallelizer is not None
    ]
    for node_id, parallelizer in node_id_non_none_parallelizer_list:
        graph.apply_parallelization_to_node(
            node_id, parallelizer, fileIdGen, fan_out, r_split_batch_size
        )


def split_hdfs_cat_input(hdfs_cat, next_node, graph, fileIdGen):
    """
    Replaces hdfs cat with a cat per block, each cat uses has an HDFSResource input fid
    Returns: A normal Cat that merges the blocks (will be removed when parallizing next_node)
    """
    assert isinstance(hdfs_cat, HDFSCat)

    ## At the moment this only works for nodes that have one standard input.
    if len(next_node.get_standard_inputs()) != 1:
        return

    hdfscat_input_id = hdfs_cat.get_standard_inputs()[0]
    hdfs_fid = graph.get_edge_fid(hdfscat_input_id)
    hdfs_filepath = str(hdfs_fid.get_resource())
    output_ids = []

    # Create a cat command per file block
    file_config = hdfs_utils.get_file_config(hdfs_filepath)
    dummy_config_path = ptempfile()  # Dummy config file, should be updated by workers
    for split_num, block in enumerate(file_config.blocks):
        resource = DFSSplitResource(
            file_config.dumps(), dummy_config_path, split_num, block.hosts
        )
        block_fid = fileIdGen.next_file_id()
        block_fid.set_resource(resource)
        graph.add_edge(block_fid)

        output_fid = fileIdGen.next_file_id()
        output_fid.make_ephemeral()
        output_ids.append(output_fid.get_ident())
        graph.add_edge(output_fid)

        split_reader_node = dfs_split_reader.make_dfs_split_reader_node(
            [block_fid.get_ident()],
            output_fid.get_ident(),
            split_num,
            config.HDFS_PREFIX,
        )
        graph.add_node(split_reader_node)

    # Remove the HDFS Cat command as it's not used anymore
    graph.remove_node(hdfs_cat.get_id())

    ## input of next command is output of new merger.
    input_id = next_node.get_standard_inputs()[0]
    new_merger = make_cat_node(output_ids, input_id)
    graph.add_node(new_merger)

    return new_merger


## This functions adds an eager on a given edge.
def add_eager(eager_input_id, graph, fileIdGen):
    new_fid = fileIdGen.next_ephemeral_file_id()
    new_id = new_fid.get_ident()

    ## TODO: seperate to better use dgsh-tee params and maybe deprecate eager
    eager_node = dgsh_tee.make_dgsh_tee_node(eager_input_id, new_id)

    ## Add the edges and the nodes to the graph
    graph.add_edge(new_fid)

    ## Modify the next node inputs to be the new inputs
    next_node_id = graph.edges[eager_input_id][2]
    if not next_node_id is None:
        next_node = graph.get_node(next_node_id)
        next_node.replace_edge(eager_input_id, new_id)
        graph.set_edge_to(new_id, next_node_id)

    graph.add_node(eager_node)


## This function adds eager nodes wherever the width of graph is
## becoming smaller.
def add_eager_nodes(graph):
    source_node_ids = graph.source_nodes()

    ## Generate a fileIdGen that doesnt clash with graph fids.
    fileIdGen = graph.get_file_id_gen()

    ## Get the next nodes
    workset = [
        node
        for source_node_id in source_node_ids
        for node in graph.get_next_nodes(source_node_id)
    ]
    visited = set()
    while len(workset) > 0:
        curr_id = workset.pop(0)
        curr = graph.get_node(curr_id)
        if not curr_id in visited:
            visited.add(curr_id)
            next_node_ids = graph.get_next_nodes(curr_id)
            workset += next_node_ids

            ## TODO: Make sure that we don't add duplicate eager nodes

            ## Add eager nodes if the node has more than one input
            curr_input_ids = graph.get_node_input_ids(curr_id)
            if len(curr_input_ids) > 1:
                ## TODO: If we know that a command reads its inputs in a list,
                ##       then we might not need to put an eager on its first input.
                ## Note: This cannot be done for `sort -m` so we need to know in the
                ##       annotations whether input consumption is in order or not.

                for curr_input_id in curr_input_ids:
                    _fid, from_node, to_node = graph.edges[curr_input_id]
                    assert to_node == curr_id
                    ## If the edge is an input edge, then we don't want to put eager.
                    if not from_node is None:
                        add_eager(curr_input_id, graph, fileIdGen)

            if isinstance(curr, Split):
                eager_input_ids = curr.get_output_list()[:-1]
                for edge_id in eager_input_ids:
                    add_eager(edge_id, graph, fileIdGen)

            ## Add an eager after r_unwrap
            if isinstance(curr, r_unwrap.RUnwrap):
                eager_input_id = curr.get_output_list()[0]
                add_eager(eager_input_id, graph, fileIdGen)

            ## Add an eager after r_split
            if isinstance(curr, r_split.RSplit):
                eager_input_ids = curr.get_output_list()
                for edge_id in eager_input_ids:
                    add_eager(edge_id, graph, fileIdGen)

    return graph



import os

import config
from ir_defs.dfg_node import DFGNode
from ir_defs.arg import Arg


class RemotePipe(DFGNode):
    def __init__(
        self,
        inputs,
        outputs,
        com_name,
        com_category,
        com_options=None,
        com_redirs=None,
        com_assignments=None,
    ):
        com_options = [] if com_options is None else com_options
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments

        super().__init__(
            inputs,
            outputs,
            com_name,
            com_category,
            com_options=com_options,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
        )


def make_remote_pipe(inputs, outputs, host_ip, port, is_remote_read, id):
    com_category = "pure"
    options = []
    opt_count = 0

    if is_remote_read:
        remote_pipe_bin = os.path.join(
            config.PASH_TOP, config.config["runtime"]["remote_read_binary"]
        )
    else:
        remote_pipe_bin = os.path.join(
            config.PASH_TOP, config.config["runtime"]["remote_write_binary"]
        )

    com_name = Arg.string_to_arg(remote_pipe_bin)

    options.append((opt_count, Arg.string_to_arg(f"--addr {host_ip}:{port}")))
    options.append((opt_count + 1, Arg.string_to_arg(f"--id {id}")))

    return RemotePipe(inputs, outputs, com_name, com_category, com_options=options)



import os
import subprocess

from shasta.ast_node import (
    AstNode,
    ArgChar,
    CArgChar,
    EArgChar,
    BArgChar,
    QArgChar,
    PipeNode,
    CommandNode,
    BackgroundNode,
)
from shasta.ast_walker import walk_ast_node
from sh_expand.expand import expand_command, ExpansionState

import config
from sh_expand.bash_expand import (
    BashExpansionState,
    expand_command as expand_command_bash,
)
from ir import IR, compile_command_to_DFG
from parse import from_ast_objects_to_shell
from util import (
    log,
    get_kv,
    make_kv,
    string_to_arguments,
    UnparallelizableError,
)


BASH_MODE = os.environ.get("pash_shell") == "bash"
if BASH_MODE:
    # doesn't need to be kept alive accross invocations
    # but should be faster to avoid creating new processes

    # too verbose to keep on with -d 1 (TODO: Get -d 2 working here)
    BASH_EXP_STATE = BashExpansionState(temp_dir=config.PASH_TMP_PREFIX, debug=False)

##
## Compile AST -> Extended AST with IRs
##
## The preprocessor sends only dataflow region nodes to the compiler:
## - PipeNode
## - CommandNode (with arguments)
## - BackgroundNode
##
## All other node types (control flow, wrappers, etc.) stay in the
## preprocessed script and are executed by bash directly.
##


def compile_asts(ast_objects: "list[AstNode]", fileIdGen, config):
    compiled_asts = []
    acc_ir = None
    for i, ast_object in enumerate(ast_objects):
        assert isinstance(ast_object, AstNode)

        ## Compile subtrees of the AST to our intermediate representation
        if BASH_MODE:
            if not BASH_EXP_STATE.is_open:
                BASH_EXP_STATE.open()
            expanded_ast = expand_command_bash(
                ast_object,
                BASH_EXP_STATE,
                config["shell_variables_file_path"],
                config["shell_variables"],
            )
        else:
            exp_state = ExpansionState(config["shell_variables"])
            expanded_ast = expand_command(ast_object, exp_state)
        compiled_ast = compile_node(expanded_ast, fileIdGen, config)

        ## If the accumulator contains an IR (meaning that the
        ## previous commands where run in background), union it with
        ## the current returned ast.
        if not acc_ir is None:
            if isinstance(compiled_ast, IR):
                acc_ir.background_union(compiled_ast)
            else:
                ## shouldn't happen since compile_node should have already
                ## raised this error
                raise UnparallelizableError(f"Node: {compiled_ast} is not pure")

            ## If the current compiled ast not in background (and so
            ## the union isn't in background too), stop accumulating
            if not acc_ir.is_in_background():
                compiled_asts.append(acc_ir)
                acc_ir = None
        else:
            ## If the compiled ast is in background, start
            ## accumulating it
            if compiled_ast.is_in_background():
                acc_ir = compiled_ast
            else:
                compiled_asts.append(compiled_ast)

    ## The final accumulator
    if not acc_ir is None:
        compiled_asts.append(acc_ir)

    return compiled_asts


def make_compile_transform(fileIdGen, config):
    """
    Create a compile transformation function for use with walk_ast_node.

    Only handles node types that actually reach the compiler from the preprocessor:
    - PipeNode, CommandNode, BackgroundNode (dataflow region nodes)
    - BArgChar, QArgChar (for command substitution and quoted strings)
    """

    def compile_transform(node):
        match node:
            case PipeNode():
                # Compile pipe children and combine into single IR
                compiled_items = [compile_node(item, fileIdGen, config) for item in node.items]
                compiled_ir = combine_pipe(compiled_items)
                compiled_ir.set_ast(node.json())
                compiled_ir.set_background(node.is_background)
                return compiled_ir

            case CommandNode():
                return compile_command_node(node, fileIdGen, config)

            case BackgroundNode():
                compiled_inner = compile_node(node.node, fileIdGen, config)
                if isinstance(compiled_inner, IR):
                    compiled_inner.set_background(True)
                    return compiled_inner
                else:
                    # Note: background nodes can be added in the distributed graph
                    # similarly to the children of pipelines.
                    node.node = compiled_inner
                    return node

            case BArgChar():
                # Command substitution - compile the inner node
                return BArgChar(node=compile_node(node.node, fileIdGen, config))

            case QArgChar():
                # Quoted string - compile the inner argument
                return QArgChar(arg=compile_command_argument(node.arg, fileIdGen, config))

            case _:
                # Let walker handle default recursion for other nodes
                return None

    return compile_transform


def compile_node(ast_object, fileIdGen, config) -> IR:
    """
    Compile an AST node to an IR or transformed AST.

    Uses walk_ast_node with a compile transformation function that handles
    the IR-specific logic for each node type.
    """
    transform = make_compile_transform(fileIdGen, config)
    return walk_ast_node(ast_object, replace=transform)


def combine_pipe(compiled_items):
    """
    Combine all compiled children of a Pipeline into a single IR.

    All children must be IRs; raises UnparallelizableError otherwise.
    """
    if not isinstance(compiled_items[0], IR):
        log("Node: {} is not pure".format(compiled_items[0]))
        raise UnparallelizableError("Node: {} is not a pure node in pipe".format(compiled_items[0]))

    combined = compiled_items[0]
    for item in compiled_items[1:]:
        if isinstance(item, IR):
            combined.pipe_append(item)
        else:
            log("Node: {} is not pure".format(item))
            raise UnparallelizableError("This specific node: {} is not a pure node in pipe".format(item))

    return combined


def compile_command_node(ast_node, fileIdGen, config):
    """
    Compile a CommandNode to IR or transformed AST.

    Tries to compile to a DFG; if that fails, returns the command
    wrapped in make_kv format.
    """
    # Compile assignments and redirection list
    compiled_assignments = compile_assignments(ast_node.assignments, fileIdGen, config)
    compiled_redirections = compile_redirections(ast_node.redir_list, fileIdGen, config)

    # This should never be possible since the preprocessor
    # wouldn't replace a call without arguments (simple assignment).
    assert len(ast_node.arguments) > 0

    arguments = ast_node.arguments
    command_name = arguments[0]
    options = compile_command_arguments(arguments[1:], fileIdGen, config)

    try:
        # If the command is not compileable to a DFG the following call will fail
        ir = compile_command_to_DFG(
            fileIdGen, command_name, options, redirections=compiled_redirections
        )
        return ir
    except ValueError as err:
        log("Command not compiled to DFG:", err)
        compiled_arguments = compile_command_arguments(arguments, fileIdGen, config)
        return make_kv(
            type(ast_node).NodeName,
            [
                ast_node.line_number,
                compiled_assignments,
                compiled_arguments,
                compiled_redirections,
            ],
        )


## This function checks if we should expand an arg_char
def should_expand_arg_char(arg_char):
    key, val = get_kv(arg_char)
    if key in ["V"]:  # Variable
        return True
    elif key == "Q":
        return should_expand_argument(val)
    elif key == "B":
        log("Cannot expand:", arg_char)
        raise NotImplementedError()
    return False


def should_expand_argument(argument):
    return any([should_expand_arg_char(arg_char) for arg_char in argument])


def execute_shell_asts(asts):
    output_script = from_ast_objects_to_shell(asts)
    exec_obj = subprocess.run(
        ["/usr/bin/env", "bash"],
        input=output_script,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    exec_obj.check_returncode()
    return exec_obj.stdout


def parse_string_to_arguments(arg_char_string):
    return string_to_arguments(arg_char_string)


def compile_arg_char(arg_char: ArgChar, fileIdGen, config):
    """Compile an arg char by recursing if it contains quotes or command substitution."""
    if isinstance(arg_char, CArgChar) or isinstance(arg_char, EArgChar):
        return arg_char
    elif isinstance(arg_char, BArgChar):
        arg_char.node = compile_node(arg_char.node, fileIdGen, config)
        return arg_char
    elif isinstance(arg_char, QArgChar):
        arg_char.arg = compile_command_argument(arg_char.arg, fileIdGen, config)
        return arg_char
    else:
        log(f"Unknown arg_char: {arg_char}")
        return arg_char


def compile_command_argument(argument, fileIdGen, config):
    return [compile_arg_char(char, fileIdGen, config) for char in argument]


def compile_command_arguments(arguments, fileIdGen, config):
    return [compile_command_argument(arg, fileIdGen, config) for arg in arguments]


def compile_assignment(assignment, fileIdGen, config):
    assignment.val = compile_command_argument(assignment.val, fileIdGen, config)
    return assignment


def compile_assignments(assignments, fileIdGen, config):
    return [compile_assignment(assignment, fileIdGen, config) for assignment in assignments]


def compile_redirection(redirection, fileIdGen, config):
    if redirection.arg is not None:
        redirection.arg = compile_command_argument(redirection.arg, fileIdGen, config)
    return redirection


def compile_redirections(redirections, fileIdGen, config):
    return [compile_redirection(redirection, fileIdGen, config) for redirection in redirections]



"""
Transformation options and state classes for AST preprocessing.
"""

from abc import ABC, abstractmethod
from enum import Enum
import os
import pickle

from shasta.ast_node import AstNode
from shasta.json_to_ast import to_ast_node

from shell_ast.ast_util import string_to_argument, make_command
from parse import from_ast_objects_to_shell
from speculative import util_spec
from util import ptempfile


# Runtime executable path - constructed from PASH_TOP environment variable
PASH_TOP = os.environ.get("PASH_TOP", "")
RUNTIME_EXECUTABLE = os.path.join(PASH_TOP, "jit_runtime/jit.sh")


class TransformationType(Enum):
    """Types of AST-to-AST transformations."""

    PASH = "pash"
    SPECULATIVE = "spec"
    AIRFLOW = "airflow"


class AbstractTransformationState(ABC):
    """Base class for transformation state."""

    def __init__(self):
        self._node_counter = 0
        self._loop_counter = 0
        self._loop_contexts = []

    def get_mode(self):
        return TransformationType.PASH

    # Node id related
    def get_next_id(self):
        new_id = self._node_counter
        self._node_counter += 1
        return new_id

    def get_current_id(self):
        return self._node_counter - 1

    def get_number_of_ids(self):
        return self._node_counter

    # Loop id related
    def get_next_loop_id(self):
        new_id = self._loop_counter
        self._loop_counter += 1
        return new_id

    def get_current_loop_context(self):
        # Return a copy
        return self._loop_contexts[:]

    def get_current_loop_id(self):
        if len(self._loop_contexts) == 0:
            return None
        else:
            return self._loop_contexts[0]

    def enter_loop(self):
        new_loop_id = self.get_next_loop_id()
        self._loop_contexts.insert(0, new_loop_id)
        return new_loop_id

    def exit_loop(self):
        self._loop_contexts.pop(0)

    @abstractmethod
    def replace_df_region(
        self, asts, disable_parallel_pipelines=False, ast_text=None
    ) -> AstNode:
        pass


class TransformationState(AbstractTransformationState):
    """Standard PaSh transformation state."""

    def replace_df_region(
        self, asts, disable_parallel_pipelines=False, ast_text=None
    ) -> AstNode:
        ir_filename = ptempfile()

        # Serialize the node in a file
        with open(ir_filename, "wb") as ir_file:
            pickle.dump(asts, ir_file)

        # Serialize the candidate df_region asts back to shell
        # so that the sequential script can be run in parallel to the compilation.
        sequential_script_file_name = ptempfile()
        text_to_output = get_shell_from_ast(asts, ast_text=ast_text)
        with open(sequential_script_file_name, "w", encoding="utf-8") as script_file:
            script_file.write(text_to_output)
        replaced_node = TransformationState.make_call_to_pash_runtime(
            ir_filename, sequential_script_file_name, disable_parallel_pipelines
        )

        return to_ast_node(replaced_node)

    @staticmethod
    def make_call_to_pash_runtime(
        ir_filename, sequential_script_file_name, disable_parallel_pipelines
    ) -> AstNode:
        """
        Make a command that calls the pash runtime with the IR file.
        """
        if disable_parallel_pipelines:
            assignments = [["pash_disable_parallel_pipelines", string_to_argument("1")]]
        else:
            assignments = [["pash_disable_parallel_pipelines", string_to_argument("0")]]
        assignments.append(
            [
                "pash_sequential_script_file",
                string_to_argument(sequential_script_file_name),
            ]
        )
        assignments.append(["pash_input_ir_file", string_to_argument(ir_filename)])

        # Call the runtime
        arguments = [
            string_to_argument("source"),
            string_to_argument(RUNTIME_EXECUTABLE),
        ]
        runtime_node = make_command(arguments, assignments=assignments)
        return runtime_node


class SpeculativeTransformationState(AbstractTransformationState):
    """Speculative execution transformation state."""

    def __init__(self, po_file: str):
        super().__init__()
        self.partial_order_file = po_file
        self.partial_order_edges = []
        self.partial_order_node_loop_contexts = {}

    def replace_df_region(
        self, asts, disable_parallel_pipelines=False, ast_text=None
    ) -> AstNode:
        text_to_output = get_shell_from_ast(asts, ast_text=ast_text)
        # Generate an ID
        df_region_id = self.get_next_id()

        # Get the current loop id and save it so that the runtime knows
        # which loop it is in.
        loop_id = self.get_current_loop_id()

        # Determine its predecessors
        if df_region_id == 0:
            predecessors = []
        else:
            predecessors = [df_region_id - 1]
        # Write to a file indexed by its ID
        util_spec.save_df_region(text_to_output, self, df_region_id, predecessors)
        replaced_node = SpeculativeTransformationState.make_call_to_spec_runtime(
            df_region_id, loop_id
        )
        return to_ast_node(replaced_node)

    def get_partial_order_file(self):
        return self.partial_order_file

    def add_edge(self, from_id: int, to_id: int):
        self.partial_order_edges.append((from_id, to_id))

    def get_all_edges(self):
        return self.partial_order_edges

    def add_node_loop_context(self, node_id: int, loop_contexts):
        self.partial_order_node_loop_contexts[node_id] = loop_contexts

    def get_all_loop_contexts(self):
        return self.partial_order_node_loop_contexts

    @staticmethod
    def make_call_to_spec_runtime(command_id: int, loop_id) -> AstNode:
        """Make a call to the speculative runtime."""
        assignments = [["pash_spec_command_id", string_to_argument(str(command_id))]]
        if loop_id is None:
            loop_id_str = ""
        else:
            loop_id_str = str(loop_id)

        assignments.append(["pash_spec_loop_id", string_to_argument(loop_id_str)])

        # Call the runtime
        arguments = [
            string_to_argument("source"),
            string_to_argument(RUNTIME_EXECUTABLE),
        ]
        runtime_node = make_command(arguments, assignments=assignments)

        return runtime_node


class AirflowTransformationState(TransformationState):
    """Airflow transformation state (same as standard PaSh for now)."""

    pass


def get_shell_from_ast(asts, ast_text=None) -> str:
    """Get shell text from AST, using original text if available."""
    if ast_text is None:
        text_to_output = from_ast_objects_to_shell(asts)
    else:
        text_to_output = ast_text
    return text_to_output



from pash_annotations.datatypes.AccessKind import (
    AccessKind,
    make_stream_output,
    make_stream_input,
    make_other_output,
)
from pash_annotations.datatypes.CommandInvocationWithIOVars import (
    CommandInvocationWithIOVars,
)

from ir_defs.dfg_node import DFGNode


class Eager(DFGNode):
    def __init__(self, cmd_invocation_with_io_vars, com_redirs=None, com_assignments=None):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments

        super().__init__(
            cmd_invocation_with_io_vars,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
        )


def make_eager_node(input_id, output_id, intermediate_file_id, eager_exec_path):
    eager_name = eager_exec_path
    intermediate_file_id_id = intermediate_file_id.get_ident()
    operand_list = [input_id, output_id, intermediate_file_id_id]
    access_map = {
        output_id: make_stream_output(),
        input_id: make_stream_input(),
        intermediate_file_id_id: make_other_output(),
    }
    cmd_inv_with_io_vars = CommandInvocationWithIOVars(
        cmd_name=eager_name,
        flag_option_list=[],
        operand_list=operand_list,
        implicit_use_of_streaming_input=None,
        implicit_use_of_streaming_output=None,
        access_map=access_map,
    )
    return Eager(cmd_inv_with_io_vars)



import os
import subprocess
import sys
from collections import namedtuple
import json
from typing import List, Tuple

HDFSBlock = namedtuple("HDFSBlock", "path hosts")


class FileData(object):
    def __init__(self, filename):
        self.blocknames = []
        self.dnodenames = []
        self.machines = []
        self.size = 0
        self.filename = filename
        # self.dnodepath = subprocess.check_output("hdfs getconf -confKey dfs.datanode.data.dir", shell=True).decode("utf-8").strip("\n")
        # if (self.dnodepath.startswith("file://")):
        #    self.dnodepath = self.dnodepath[len("file://"):]

    def paths(self):
        assert len(self.blocknames) != 0
        assert len(self.dnodenames) != 0
        assert self.size > 0
        filepaths = []
        for i in range(len(self.blocknames)):
            filepaths.append(
                os.path.join(
                    "current",
                    self.dnodenames[i],
                    "current",
                    "finalized",
                    "subdir0",
                    "subdir0",
                    self.blocknames[i],
                )
            )
        return filepaths


class HDFSFileConfig:
    def __init__(self, filedata: FileData):
        self.blocks: List[HDFSBlock] = []
        for i, block_path in enumerate(filedata.paths()):
            hosts = list(map(lambda addr: addr.rsplit(":", 1)[0], filedata.machines[i]))
            self.blocks.append(HDFSBlock(block_path, hosts))

    def _serialize(self):
        data = {"blocks": []}
        for path, hosts in self.blocks:
            data["blocks"].append({"path": path, "hosts": hosts})
        return data

    def dumps(self):
        data = self._serialize()
        return json.dumps(data)

    def dump(self, filepath):
        data = self._serialize()
        with open(filepath, "w") as f:
            json.dump(data, f)

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, HDFSFileConfig):
            return False
        return self.blocks == __o.blocks


def get_hdfs_file_data(filename):
    info = FileData(filename)
    log = subprocess.check_output(
        "hdfs fsck {0} -files -blocks -locations".format(filename),
        shell=True,
        stderr=subprocess.PIPE,
    )
    count = 0
    for line in log.splitlines():
        wordarr = line.split()
        if len(wordarr) > 0 and wordarr[0].decode("utf-8") == filename and count == 0:
            info.size = int(wordarr[1])
            count += 1
        elif (
            len(wordarr) > 0
            and count > 0
            and wordarr[0][:-1].isdigit()
            and int(wordarr[0][:-1]) == count - 1
        ):
            count += 1
            rawinfo = wordarr[1].decode("utf-8").split(":")
            info.blocknames.append(rawinfo[1][0 : rawinfo[1].rfind("_")])
            info.dnodenames.append(rawinfo[0])
            stline = line.decode("utf-8")
            info.machines.append(
                _getIPs(stline[stline.find("DatanodeInfoWithStorage") - 1 :])
            )
    assert len(info.blocknames) != 0
    assert len(info.dnodenames) != 0
    assert info.size > 0
    return info


def _getIPs(raw):
    rawparts = raw.split(" ")
    ips = []
    for part in rawparts:
        index = part.find("DatanodeInfoWithStorage")
        ips.append(part[index + len("DatanodeInfoWithStorage") + 1 : part.find(",")])
    return ips


if __name__ == "__main__":
    assert len(sys.argv) == 2
    filename = sys.argv[1]
    info = get_hdfs_file_data(filename)
    print("Size = ", info.size)
    paths = info.paths()
    for i in range(len(paths)):
        print("Machines = ", info.machines[i])
        print("Block = ", paths[i])



import copy
import os

from ir_defs.redirection import Redirection
from ir_defs.resource import FileResource

from annotations_utils.util_cmd_invocations import (
    to_node_cmd_inv_with_io_vars,
    construct_property_container_from_list_of_properties,
)

from util import (
    log,
    return_empty_list_if_none_else_itself,
    return_default_if_none_else_itself,
)


## Assumption: Everything related to a DFGNode must be already expanded.
## TODO: Ensure that this is true with assertions
class DFGNode:
    ## Unique identifier for nodes
    next_id = 0

    ## cmd_invocation_with_io_vars : command invocation data structure with edge ids as symbolic variables for filenames etc.
    ## com_redirs : list of redirections
    ## com_assignments : list of assignments
    ## parallelizer_list : list of parallelizers for this DFGNode
    ## cmd_related_properties : dict to store properties like commutativity
    def __init__(
        self,
        cmd_invocation_with_io_vars,
        com_redirs=None,
        com_assignments=None,
        parallelizer_list=None,
        cmd_related_properties=None,
    ):
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments

        ## @KK: can this be deleted? Was there another id in the member attributes before?
        ## Add a unique identifier to each DFGNode since id() is not guaranteed to be unique for objects that have different lifetimes.
        ## This leads to issues when nodes are deleted and new ones are created, leading to id() clashes between them
        self.id = DFGNode.next_id
        DFGNode.next_id += 1

        self.com_redirs = [Redirection(redirection) for redirection in com_redirs]
        self.com_assignments = com_assignments
        self.parallelizer_list = return_empty_list_if_none_else_itself(
            parallelizer_list
        )
        default_cmd_properties = construct_property_container_from_list_of_properties(
            []
        )
        self.cmd_related_properties = return_default_if_none_else_itself(
            cmd_related_properties, default_cmd_properties
        )
        self.cmd_invocation_with_io_vars = cmd_invocation_with_io_vars
        # log("Node created:", self.id, self)

    def __repr__(self):
        # TODO: add other attributes
        return str(self.cmd_invocation_with_io_vars)

    ## Generates a dot node for the DFG node
    def add_dot_node(self, dot, node_id):
        label = self.get_dot_label()
        dot.node(str(node_id), label=label)
        return dot

    ## Get the label of the node. By default, it is simply the name
    def get_dot_label(self) -> str:
        ## The name could be a full path
        name = self.cmd_invocation_with_io_vars.cmd_name
        basename = os.path.basename(str(name))
        return basename

    def get_id(self):
        return self.id

    ## Copying requires setting the id to a new one too
    def copy(self):
        node_copy = copy.deepcopy(self)
        node_copy.id = DFGNode.next_id
        DFGNode.next_id += 1
        return node_copy

    def get_input_list(self):
        inputs = self.cmd_invocation_with_io_vars.generate_inputs()
        return inputs.get_all_inputs()

    def get_output_list(self):
        return self.cmd_invocation_with_io_vars.generate_outputs()

    def get_streaming_inputs(self):
        inputs = self.cmd_invocation_with_io_vars.generate_inputs()
        return inputs.get_streaming_inputs()

    def get_configuration_inputs(self):
        inputs = self.cmd_invocation_with_io_vars.generate_inputs()
        return inputs.get_config_inputs()

    def is_commutative(self):
        val = self.cmd_related_properties.get_property_value("is_commutative")
        if val is not None:
            return val
        else:
            return False

    ## Auxiliary method that returns any necessary redirections,
    ##   at the moment it doesn't look necessary.
    def _to_ast_aux_get_redirs(self):
        ## still used in to_ast
        ## TODO: Properly handle redirections
        ##
        ## TODO: If one of the redirected outputs or inputs is changed in the IR
        ##       (e.g. `cat < s1` was changed to read from an ephemeral file `cat < "#file5"`)
        ##       this needs to be changed in the redirections too. Maybe we can modify redirections
        ##       when replacing fid.
        ##
        ## It seems that if we have already applied redirections we might not need to
        ## care about them anymore (since they will be created in new_redirs.)
        ##
        ## redirs = [redir_defs.to_ast() for redir in self.com_redirs]
        ##
        ## At the moment we do not reprint redirections here (we only produce redirections
        ## where we recreate arguments and redirections).
        return []

    ## TODO: Improve this function to be separately implemented for different special nodes,
    ##       such as cat, eager, split, etc...
    ## I do not think this separation is reasonable anymore since we remodelled nodes in a way that the back-translation is trivial
    ## One exception:
    ##  - r_wrap; currently, the wrapped command is translated at creation of the r_wrap already and
    ##    hence assumes that non-streaming inputs/outputs will not change; with a special to_ast, we could circumvent this
    def to_ast(self, edges, drain_streams):
        ## TODO: We might not want to implement this at all actually
        if drain_streams:
            raise NotImplementedError()
        else:
            # commented since "see above"
            ## Handle special node to ast here
            # node = self.special_to_ast(edges)
            # if node is not None:
            #     return node

            redirs = self._to_ast_aux_get_redirs()
            assignments = self.com_assignments

            node = to_node_cmd_inv_with_io_vars(
                self.cmd_invocation_with_io_vars, edges, redirs, assignments
            )
            # TODO: think about redirections
            # old code for this:
            # rest_argument_fids, new_redirs = create_command_arguments_redirs(com_name_ast,
            #                                                                  option_asts,
            #                                                                  input_fids,
            #                                                                  output_fids)
        return node

    ## This method applies the redirections to get the correct, inputs, outputs of a node.
    ##
    ## WARNING: For now it only works with 'To' redirections for
    ## stdout, and it applies them by adding a resource to the stdout
    ## of the command. It also keeps them for possible future usage.
    ##
    ## TODO: Properly handle all redirections. This requires a nice
    ## abstraction. Maybe the best way would be to keep them around
    ## and always recompute inputs/outputs when needed by following
    ## the redirections.
    ##
    ## TODO: Is it correct to apply redirections from left to right?
    def apply_redirections(self, edges):
        unhandled_redirs = []
        for redirection in self.com_redirs:
            ## Handle To redirections that have to do with stdout
            if redirection.is_to_file() and redirection.is_for_stdout():
                # log(redirection)
                file_resource = FileResource(redirection.file_arg)
                success = False
                for i in range(len(self.get_output_list())):
                    output_edge_id = self.get_output_list()[i]
                    output_fid = edges[output_edge_id][0]
                    if (
                        output_fid.has_file_descriptor_resource()
                        and output_fid.resource.is_stdout()
                    ):
                        success = True
                        edges[output_edge_id][0].set_resource(file_resource)
                        # self.outputs[i].set_resource(file_resource)
                assert success
            elif redirection.is_from_file() and redirection.is_for_stdin():
                # log(redirection)
                file_resource = FileResource(redirection.file_arg)
                success = False
                for input_edge_id in self.get_input_list():
                    input_fid = edges[input_edge_id][0]
                    if (
                        input_fid.has_file_descriptor_resource()
                        and input_fid.resource.is_stdin()
                    ):
                        success = True
                        edges[input_edge_id][0].set_resource(file_resource)
                assert success
            else:
                log("Warning -- Unhandled redirection:", redirection)
                unhandled_redirs.append(redirection)
                ## TODO: I am not sure if this is the correct way to handle unhandled redirections.
                ##       Does it make any sense to keep them and have them in the Final AST.
                raise NotImplementedError()

    ## This renames the from_id (wherever it exists in inputs or outputs)
    ## to the to_id.
    ##
    ## TODO: Make sure we don't need to change redirections here.
    ##
    ## TODO: Make this a method of graph to change the from, to too.
    def replace_edge(self, from_id, to_id):
        self.cmd_invocation_with_io_vars.replace_var(from_id, to_id)

    ## TODO: There must be a lib function to do this.
    def replace_edge_in_list(self, edge_ids, from_id, to_id):
        new_edge_ids = []
        for id in edge_ids:
            if id == from_id:
                new_edge_id = to_id
            else:
                new_edge_id = id
            new_edge_ids.append(new_edge_id)
        return new_edge_ids

    def get_option_implemented_round_robin_parallelizer(self):
        for parallelizer in self.parallelizer_list:
            splitter = parallelizer.get_splitter()
            if (
                splitter.is_splitter_round_robin()
                and parallelizer.are_all_parts_implemented()
            ):
                return parallelizer
        return None

    def get_option_implemented_round_robin_with_unwrap_parallelizer(self):
        for parallelizer in self.parallelizer_list:
            splitter = parallelizer.get_splitter()
            if (
                splitter.is_splitter_round_robin_with_unwrap_flag()
                and parallelizer.are_all_parts_implemented()
            ):
                return parallelizer
        return None

    def get_option_implemented_consecutive_chunks_parallelizer(self):
        for parallelizer in self.parallelizer_list:
            splitter = parallelizer.get_splitter()
            if (
                splitter.is_splitter_consec_chunks()
                and parallelizer.are_all_parts_implemented()
            ):
                return parallelizer
        return None

    @staticmethod
    def make_simple_dfg_node_from_cmd_inv_with_io_vars(cmd_inv_with_io_vars):
        return DFGNode(cmd_inv_with_io_vars)

    def get_single_streaming_input_single_output_and_configuration_inputs_of_node_for_parallelization(
        self,
    ):
        streaming_inputs = self.get_streaming_inputs()
        assert len(streaming_inputs) == 1
        streaming_input = streaming_inputs[0]
        configuration_inputs = self.get_configuration_inputs()
        assert len(configuration_inputs) == 0
        streaming_outputs = self.get_output_list()
        assert len(streaming_outputs) == 1
        streaming_output = streaming_outputs[0]
        return streaming_input, streaming_output, configuration_inputs



from ir_defs.dfg_node import DFGNode
from ir_defs.arg import Arg

from annotations_utils.util_cmd_invocations import (
    get_command_invocation_prefix_from_dfg_node,
)
from util import log


## This class corresponds to a generic n-ary aggregator
##
## TODO: Do we need to do anything special for binary aggregators?
class MapperAggregatorNode(DFGNode):
    def __init__(
        self,
        old_node,
        input_ids,
        output_ids,
        name_string,
        new_options,
        flag_option_list,
    ):
        ## The name of the aggregator command
        name = Arg.string_to_arg(name_string)

        ## TODO: The category should also be acquired through annotations (and maybe should be asserted to be at most pure)
        com_category = "pure"

        ## TODO: Not sure if redirections need to be copied to new function.
        com_redirs = [redir_defs.to_ast() for redir in old_node.com_redirs]
        super().__init__(
            input_ids,
            output_ids,
            name,
            com_category,
            com_options=new_options,  # changed that all are already in there and not appended
            flag_option_list=flag_option_list,
            com_redirs=com_redirs,
            com_assignments=old_node.com_assignments,
        )


class AggregatorNode(MapperAggregatorNode):
    def __init__(self, old_node, input_ids, output_ids):
        used_parallelizer = old_node.get_used_parallelizer()
        cmd_inv_pref = get_command_invocation_prefix_from_dfg_node(old_node)
        used_aggregator = used_parallelizer.get_actual_aggregator(cmd_inv_pref)
        log(f"used_agg: {used_aggregator}")
        log(f"old_node: {old_node}")

        ## Check if an aggregator can be instantiated from the node
        if used_aggregator is None:
            log(
                "Error: Node:",
                old_node,
                "does not contain information to instantiate an aggregator!",
            )
            raise Exception("No information to instantiate aggregator")

        ## The name of the aggregator command
        agg_name_string = used_aggregator.cmd_name
        all_options_incl_new = [
            Arg.string_to_arg(el.get_name())
            for el in used_aggregator.flag_option_list
            + used_aggregator.positional_config_list
        ]
        # TODO: zip is nicer
        all_options_incl_new_right_format = [
            (i, all_options_incl_new[i]) for i in range(len(all_options_incl_new))
        ]

        super().__init__(
            old_node,
            input_ids,
            output_ids,
            agg_name_string,
            all_options_incl_new_right_format,
            flag_option_list=used_aggregator.flag_option_list,
        )

        log("Generic Aggregator Created:", self)



from ir_defs.dfg_node import DFGNode


class HDFSCat(DFGNode):
    def __init__(
        self,
        inputs,
        outputs,
        com_name,
        com_category,
        com_options=None,
        com_redirs=None,
        com_assignments=None,
    ):
        com_options = [] if com_options is None else com_options
        com_redirs = [] if com_redirs is None else com_redirs
        com_assignments = [] if com_assignments is None else com_assignments

        assert str(com_name) == "hdfs"
        assert str(com_options[0][1]) == "dfs" and str(com_options[1][1]) == "-cat"
        super().__init__(
            inputs,
            outputs,
            com_name,
            com_category,
            com_options=com_options,
            com_redirs=com_redirs,
            com_assignments=com_assignments,
        )



from shasta.ast_node import RedirectionNode, FileRedirNode
from ir_defs.arg import Arg
from util import make_kv, UnparallelizableError

class Redirection:
    def __init__(self, redirection: RedirectionNode):
        ## TODO: Support all redirections
        if not isinstance(redirection, FileRedirNode) or redirection.redir_type not in [
            "To",
            "From",
        ]:
            raise NotImplementedError(redirection)

        self.redir_type = redirection.NodeName
        self.redir_subtype = redirection.redir_type
        self.stream_id = redirection.fd
        self.file_arg = Arg(redirection.arg)

        # log(redirection)

    def __repr__(self):
        return "({}, {}, {}, {})".format(
            self.redir_type, self.redir_subtype, self.stream_id, self.file_arg
        )

    def to_ast(self):
        redir = make_kv(
            self.redir_type,
            [self.redir_subtype, self.stream_id, self.file_arg.to_ast()],
        )
        return redir

    def is_to_file(self):
        return self.redir_type == "File" and self.redir_subtype == "To"

    def is_for_stdout(self):
        return self.stream_id == 1

    def is_from_file(self):
        return self.redir_type == "File" and self.redir_subtype == "From"

    def is_for_stdin(self):
        return self.stream_id == 0



"""
Test cases for walk_preprocess module.

This module tests the refactored preprocessing walker against various
shell script constructs to ensure correct behavior.
"""

import os
import sys
import tempfile
import unittest

# Add compiler directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# Set required environment variables before importing modules
os.environ.setdefault("PASH_TOP", os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))
os.environ.setdefault("PASH_TMP_PREFIX", tempfile.mkdtemp() + "/")
os.environ.setdefault("PASH_BASH_VERSION", "5 2 21")

from parse import parse_shell_to_asts
from shell_ast.transformation_options import TransformationState
from shell_ast.ast_to_ast import preprocess_node
from shell_ast.walk_preprocess import WalkPreprocess, PreprocessContext
from pash_preprocessor import preprocess


def preprocess_close_node(ast_node, trans_options, last_object=False):
    """Preprocess a node with close-node semantics (for backward compatibility with tests)."""
    from shell_ast.handlers.loop_tracking import for_node_with_loop_tracking
    handlers = {"for": for_node_with_loop_tracking}
    walker = WalkPreprocess(handlers=handlers)
    ctx = PreprocessContext(trans_options=trans_options, last_object=last_object)
    return walker.walk_close(ast_node, ctx)


class MockArgs:
    """Mock arguments for preprocessing."""
    def __init__(self):
        self.bash = False
        self.preprocess_mode = "pash"


def parse_script(script: str):
    """Helper to parse a shell script string to AST objects."""
    with tempfile.NamedTemporaryFile(mode='w', suffix='.sh', delete=False) as f:
        f.write(script)
        f.flush()
        script_path = f.name

    try:
        return list(parse_shell_to_asts(script_path, bash_mode=False))
    finally:
        os.unlink(script_path)


def preprocess_script(script: str) -> str:
    """Helper to preprocess a script and return the result."""
    with tempfile.NamedTemporaryFile(mode='w', suffix='.sh', delete=False) as f:
        f.write(script)
        f.flush()
        script_path = f.name

    try:
        args = MockArgs()
        return preprocess(script_path, args)
    finally:
        os.unlink(script_path)


class TestPreprocessNode(unittest.TestCase):
    """Test preprocess_node function for individual node types."""

    def test_command_node(self):
        """Test that CommandNode with arguments is marked for replacement."""
        ast_objects = parse_script("echo hello")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, True)
        self.assertEqual(result.non_maximal, False)
        self.assertEqual(result.something_replaced, True)

    def test_command_assignment_only(self):
        """Test that assignment-only CommandNode is not marked for replacement."""
        ast_objects = parse_script("x=1")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, False)

    def test_pipe_node(self):
        """Test that PipeNode is marked for replacement."""
        ast_objects = parse_script("echo hello | grep h")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, True)
        self.assertEqual(result.non_maximal, False)
        self.assertEqual(result.something_replaced, True)

    def test_background_node(self):
        """Test that BackgroundNode is marked as non-maximal."""
        ast_objects = parse_script("cmd &")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, True)
        self.assertEqual(result.non_maximal, True)
        self.assertEqual(result.something_replaced, True)

    def test_semi_node(self):
        """Test that SemiNode recurses into children."""
        ast_objects = parse_script("echo hello; echo world")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_and_node(self):
        """Test that AndNode recurses into children."""
        ast_objects = parse_script("cmd1 && cmd2")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_or_node(self):
        """Test that OrNode recurses into children."""
        ast_objects = parse_script("cmd1 || cmd2")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_not_node(self):
        """Test that NotNode recurses into body."""
        ast_objects = parse_script("! cmd")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_if_node(self):
        """Test that IfNode recurses into all branches."""
        ast_objects = parse_script("if true; then echo yes; else echo no; fi")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_for_node(self):
        """Test that ForNode recurses into body."""
        ast_objects = parse_script("for x in 1 2 3; do echo $x; done")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_while_node(self):
        """Test that WhileNode recurses into test and body."""
        ast_objects = parse_script("while true; do echo x; done")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_subshell_node(self):
        """Test that SubshellNode recurses into body."""
        ast_objects = parse_script("(echo subshell)")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_group_node_as_command(self):
        """Test brace group parsing behavior.

        Note: The POSIX (dash) parser optimizes away group nodes for simple
        cases like '{ echo group; }', turning them into plain CommandNode.
        GroupNode is only produced by the bash parser.
        This test verifies we handle what we get from the parser.
        """
        # With dash parser, '{ echo group; }' parses as CommandNode
        ast_objects = parse_script("{ echo group; }")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        # The parser returns CommandNode, which should be replaced
        self.assertEqual(type(ast).__name__, "CommandNode")
        self.assertEqual(result.replace_whole, True)
        self.assertEqual(result.something_replaced, True)

    def test_case_node(self):
        """Test that CaseNode recurses into case bodies."""
        ast_objects = parse_script("case $x in a) echo a;; esac")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, True)

    def test_defun_node(self):
        """Test that DefunNode is not preprocessed (per current logic)."""
        ast_objects = parse_script("myfunc() { echo hello; }")
        self.assertEqual(len(ast_objects), 1)

        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()
        result = preprocess_node(ast, trans_options, last_object=True)

        self.assertEqual(result.replace_whole, False)
        self.assertEqual(result.something_replaced, False)


class TestFullPreprocessing(unittest.TestCase):
    """Test full preprocessing pipeline."""

    def test_simple_command(self):
        """Test preprocessing of simple command."""
        result = preprocess_script("echo hello")
        self.assertIn("jit", result)

    def test_pipeline(self):
        """Test preprocessing of pipeline."""
        result = preprocess_script("echo hello | grep h")
        self.assertIn("jit", result)

    def test_for_loop_injects_tracking(self):
        """Test that for loop preprocessing injects loop tracking."""
        result = preprocess_script("for x in 1 2 3; do echo $x; done")
        self.assertIn("pash_loop_0_iter", result)
        self.assertIn("pash_loop_iters", result)

    def test_nested_for_loops(self):
        """Test that nested for loops have different loop IDs."""
        result = preprocess_script("for x in 1 2; do for y in a b; do echo $x $y; done; done")
        self.assertIn("pash_loop_0_iter", result)
        self.assertIn("pash_loop_1_iter", result)

    def test_assignment_preserved(self):
        """Test that assignment without command is preserved as-is."""
        result = preprocess_script("x=1")
        self.assertEqual(result.strip(), "x=1")

    def test_if_statement(self):
        """Test preprocessing of if statement."""
        result = preprocess_script("if true; then echo yes; fi")
        self.assertIn("jit", result)

    def test_complex_script(self):
        """Test preprocessing of complex script."""
        script = """
echo start
for x in 1 2 3; do
    echo $x | grep 1
done
if true; then
    cat file | sort
else
    cat file | uniq
fi
echo end
"""
        result = preprocess_script(script)
        self.assertIn("jit", result)
        self.assertIn("pash_loop_0_iter", result)


class TestPreprocessCloseNode(unittest.TestCase):
    """Test preprocess_close_node function."""

    def test_close_node_replaces_command(self):
        """Test that close_node replaces command nodes.

        Note: The replacement is still a CommandNode, but it's a different
        CommandNode that sources the pash_runtime.sh script.
        """
        ast_objects = parse_script("echo hello")
        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()

        final_ast, something_replaced = preprocess_close_node(
            ast, trans_options, last_object=True
        )

        self.assertTrue(something_replaced)
        # The returned AST is still a CommandNode but with different content
        # (it's the runtime call). We verify this by checking it's different
        # from the original.
        self.assertIsNot(final_ast, ast)

    def test_close_node_does_not_replace_assignment(self):
        """Test that close_node does not replace assignment-only nodes."""
        ast_objects = parse_script("x=1")
        ast, _, _, _ = ast_objects[0]
        trans_options = TransformationState()

        final_ast, something_replaced = preprocess_close_node(
            ast, trans_options, last_object=True
        )

        self.assertFalse(something_replaced)
        # The AST should be unchanged
        self.assertEqual(type(final_ast).__name__, "CommandNode")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
from shasta.ast_node import ArgChar, CArgChar
from util import format_arg_chars, string_to_carg_char_list, log


class Arg:
    arg_char_list: "list[ArgChar]"

    def __init__(self, arg_char_list: "list[ArgChar]"):
        assert not isinstance(arg_char_list, Arg)
        for arg_char in arg_char_list:
            assert isinstance(arg_char, ArgChar)
        self.arg_char_list = arg_char_list

    def __repr__(self):
        return format_arg_chars(self.arg_char_list)

    def __eq__(self, other):
        if isinstance(other, Arg):
            return self.arg_char_list == other.arg_char_list
        log(
            "Warning: Comparing Arg:",
            self,
            "with a non Arg argument:",
            other,
            "of type:",
            type(other),
        )
        return False

    def opt_serialize(self):
        return self.__repr__()

    def to_ast(self):
        return self.arg_char_list

    def concatenate(self, other):
        space = [CArgChar(32)]  # space
        self.arg_char_list.extend(space)
        self.arg_char_list.extend(other.arg_char_list)

    @staticmethod
    def string_to_arg(string: str) -> Arg:
        return Arg(string_to_carg_char_list(string))



import os
import pickle
import socket
import sys
from collections import deque, defaultdict
from typing import List, Tuple, Dict, Callable
from uuid import uuid4

sys.path.append("/pash/compiler")

import config
import pash_compiler
from ir import IR, FileIdGen
from ir_to_ast import to_shell
from util import ptempfile
from dspash.hdfs_utils import HDFSFileConfig

from ir_defs.file_id import FileId
from ir_defs.resource import DFSSplitResource, FileDescriptorResource

import ir_defs.nodes.remote_pipe as remote_pipe

HOST = socket.gethostbyname(socket.gethostname())
NEXT_PORT = 58000
DISCOVERY_PORT = 50052


def read_graph(filename):
    with open(filename, "rb") as ir_file:
        ir, shell_vars = pickle.load(ir_file)
    return ir, shell_vars


def save_configs(graph: IR, dfs_configs_paths: Dict[HDFSFileConfig, str]):
    for edge in graph.all_fids():
        if isinstance(edge.get_resource(), DFSSplitResource):
            resource: DFSSplitResource = edge.get_resource()
            config: HDFSFileConfig = resource.config
            if config not in dfs_configs_paths:
                config_path = ptempfile()
                with open(config_path, "w") as f:
                    f.write(config)
                dfs_configs_paths[config] = config_path
            else:
                config_path = dfs_configs_paths[config]

            resource.set_config_path(config_path)


def to_shell_file(graph: IR, args) -> str:
    filename = ptempfile()

    dirs = set()
    for edge in graph.all_fids():
        directory = os.path.join(config.PASH_TMP_PREFIX, edge.prefix)
        dirs.add(directory)

    for directory in dirs:
        os.makedirs(directory, exist_ok=True)

    if not args.no_eager:
        graph = pash_compiler.add_eager_nodes(graph)

    script = to_shell(graph, args)
    with open(filename, "w", encoding="utf-8") as f:
        f.write(script)
    return filename


def split_ir(graph: IR) -> Tuple[List[IR], Dict[int, IR]]:
    r""" Takes an optimized IR and splits it subgraphs. Every subgraph
    is a continues section between a splitter and a merger.

    Example: given the following optimized IR
                      - tr -- grep -
                    /                \
    cat - uniq -split                    cat - wc
                    \                /
                      - tr -- grep -
    
    The function returns the following list of IRs:
    [cat - uniq - split, tr - grep, tr - grep, cat - wc]

    Args:
        graph: an IR optimized by the pash compiler
    Returns:
        subgraphs: a list of IRs representing sections of the original graph
        input_fifo_map: a mapping from input edge id to the subgraph which
            have that input edge. This is used later to directly traverse
            from output edge of one subgraph to corrosponding input subgraph

    """
    source_node_ids = graph.source_nodes()
    input_fifo_map = defaultdict(list)

    subgraphs = []
    queue = deque([(source, IR({}, {})) for source in source_node_ids])

    # Graph is a DAG so we need to keep track of traversed edges
    visited_edges = set(graph.all_input_fids())
    visited_nodes = set()

    while queue:
        old_node_id, subgraph = queue.popleft()
        input_fids = graph.get_node_input_fids(old_node_id)
        output_fids = graph.get_node_output_fids(old_node_id)

        if any(map(lambda fid: fid not in visited_edges, input_fids)):
            if subgraph.source_nodes():
                subgraphs.append(subgraph)
            continue

        # Second condition makes sure we don't add empty graphs
        if len(input_fids) > 1 and subgraph.source_nodes():  # merger node
            if subgraph not in subgraphs:
                subgraphs.append(subgraph)
            subgraph = IR({}, {})

        if old_node_id in visited_nodes:
            continue
        else:
            visited_nodes.add(old_node_id)

        node = graph.get_node(old_node_id).copy()
        node_id = node.get_id()

        for idx, input_fid in enumerate(input_fids):
            input_edge_id = None
            # If subgraph is empty and edge isn't ephemeral the edge needs to be added
            if not input_fid.get_ident() in subgraph.edges:
                new_fid = input_fid
                subgraph.add_to_edge(new_fid, node_id)
                input_edge_id = new_fid.get_ident()
            else:
                input_edge_id = input_fid.get_ident()
                subgraph.set_edge_to(input_edge_id, node_id)
            # keep track
            input_fifo_map[input_edge_id].append(subgraph)

        # Add edges coming out of the node
        for output_fid in output_fids:
            subgraph.add_from_edge(node_id, output_fid)
            visited_edges.add(output_fid)

        # Add edges coming into the node
        for input_fid in input_fids:
            if input_fid.get_ident() not in subgraph.edges:
                subgraph.add_to_edge(input_fid, node_id)

        # Add the node
        subgraph.add_node(node)

        next_ids = graph.get_next_nodes(old_node_id)
        if len(next_ids) == 1:
            queue.append((next_ids[0], subgraph))
        else:
            subgraphs.append(subgraph)
            for next_id in next_ids:
                queue.append((next_id, IR({}, {})))

    # print(list(map(lambda k : k.all_fids(), graphs)))
    return subgraphs, input_fifo_map


def add_stdout_fid(graph: IR, file_id_gen: FileIdGen) -> FileId:
    stdout = file_id_gen.next_file_id()
    stdout.set_resource(FileDescriptorResource(("fd", 1)))
    graph.add_edge(stdout)
    return stdout


def assign_workers_to_subgraphs(
    subgraphs: List[IR],
    file_id_gen: FileIdGen,
    input_fifo_map: Dict[int, IR],
    get_worker: Callable,
) -> (IR, Tuple):
    """Takes a list of subgraphs and assigns a worker to each subgraph and augment
    the subgraphs with the necessary remote read/write nodes for data movement
    between workers. This function also produces graph that should run in
    the original shell in which pash was executed. This graph contains
    remote read/write nodes for stdin/stdout, named pipes, and files.

    Args:
        subgraphs: list of sub sections of an optimized IR (returned from split_ir)
        file_id_gen: file id generator of the original ir
        input_fifo_map: mapping from input idge id to subgraph (returned from split_ir)
        get_worker: a callback for getting a worker from worker manager
    Returns:
        main_graph: the graph to execute on main shell
        worker_subgraph_pairs: A list of pairs representing which worker
            each subgraph should be executed on.
    """
    # The graph to execute in the main pash_compiler
    main_graph = IR({}, {})
    worker_subgraph_pairs = []

    # Replace output edges and corrosponding input edges with remote read/write
    for subgraph in subgraphs:
        subgraph_critical_fids = list(
            filter(lambda fid: fid.has_remote_file_resource(), subgraph.all_fids())
        )
        worker = get_worker(subgraph_critical_fids)
        worker._running_processes += 1
        worker_subgraph_pairs.append((worker, subgraph))
        sink_nodes = subgraph.sink_nodes()
        assert len(sink_nodes) == 1

        for out_edge in subgraph.get_node_output_fids(sink_nodes[0]):
            stdout = add_stdout_fid(subgraph, file_id_gen)
            out_edge_id = out_edge.get_ident()
            # Replace the old edge with an ephemeral edge in case it isn't and
            # to avoid modifying the edge in case it's used in some other subgraph
            ephemeral_edge = file_id_gen.next_ephemeral_file_id()
            subgraph.replace_edge(out_edge_id, ephemeral_edge)
            edge_uid = uuid4()
            # Add remote-write node at the end of the subgraph
            remote_write = remote_pipe.make_remote_pipe(
                [ephemeral_edge.get_ident()],
                [stdout.get_ident()],
                worker.host(),
                DISCOVERY_PORT,
                False,
                edge_uid,
            )
            subgraph.add_node(remote_write)

            # Copy the old output edge resource
            new_edge = file_id_gen.next_file_id()
            new_edge.set_resource(out_edge.get_resource())
            # Get the subgraph which "edge" writes to
            if out_edge_id in input_fifo_map and out_edge.is_ephemeral():
                # Copy the old output edge resource
                matching_subgraph = input_fifo_map[out_edge_id][0]
                matching_subgraph.replace_edge(out_edge.get_ident(), new_edge)
            else:
                matching_subgraph = main_graph
                matching_subgraph.add_edge(new_edge)

            remote_read = remote_pipe.make_remote_pipe(
                [],
                [new_edge.get_ident()],
                worker.host(),
                DISCOVERY_PORT,
                True,
                edge_uid,
            )
            matching_subgraph.add_node(remote_read)

    # Replace non ephemeral input edges with remote read/write
    for subgraph in subgraphs:
        source_nodes = subgraph.source_nodes()
        for source in source_nodes:
            for in_edge in subgraph.get_node_input_fids(source):
                if (
                    in_edge.has_file_resource()
                    or in_edge.has_file_descriptor_resource()
                ):
                    # setup
                    stdout = add_stdout_fid(main_graph, file_id_gen)

                    # Copy the old input edge resource
                    new_edge = file_id_gen.next_file_id()
                    new_edge.set_resource(in_edge.get_resource())
                    main_graph.add_edge(new_edge)

                    # Add remote write to main subgraph
                    edge_uid = uuid4()
                    remote_write = remote_pipe.make_remote_pipe(
                        [new_edge.get_ident()],
                        [stdout.get_ident()],
                        HOST,
                        DISCOVERY_PORT,
                        False,
                        edge_uid,
                    )
                    main_graph.add_node(remote_write)

                    # Add remote read to current subgraph
                    ephemeral_edge = file_id_gen.next_ephemeral_file_id()
                    subgraph.replace_edge(in_edge.get_ident(), ephemeral_edge)

                    remote_read = remote_pipe.make_remote_pipe(
                        [],
                        [ephemeral_edge.get_ident()],
                        HOST,
                        DISCOVERY_PORT,
                        True,
                        edge_uid,
                    )
                    subgraph.add_node(remote_read)
                else:
                    # sometimes a command can have both a file resource and an ephemeral resources (example: spell oneliner)
                    continue

    return main_graph, worker_subgraph_pairs


def prepare_graph_for_remote_exec(filename: str, get_worker: Callable):
    """
    Reads the complete ir from filename and splits it
    into subgraphs where ony the first subgraph represent a continues
    segment (merger segment or branched segment) in the graph.
    Note: All subgraphs(except first one) read and write from remote pipes.
        However, we had to add a fake stdout to avoid some problems when converting to shell code.

    Returns:
        worker_graph_pairs: List of (worker, subgraph)
        shell_vars: shell variables
        main_graph: The ir we need to execute on the main shell.
            This graph contains edges to correctly redirect the following to remote workers
            - special pipes (stdin/stdout)
            - named pipes reading and writing
            - files reading and writing
    """
    ir, shell_vars = read_graph(filename)
    file_id_gen = ir.get_file_id_gen()
    subgraphs, mapping = split_ir(ir)
    main_graph, worker_graph_pairs = assign_workers_to_subgraphs(
        subgraphs, file_id_gen, mapping, get_worker
    )
    return worker_graph_pairs, shell_vars, main_graph



"""
Argument parser classes for PaSh compiler components.

Classes:
- BaseParser: Base argparse class with common flags (debug, log_file)
- CompilerParser: Extended parser with compiler-specific arguments
"""

import argparse
import os


class BaseParser(argparse.ArgumentParser):
    """
    Base class for all Argument Parsers used by PaSh. It has two configurable flags
    by default: debug and log_file.

    Other flags are available by classes which inherit BaseParser
    """

    @staticmethod
    def _get_width():
        cpus = os.cpu_count()
        assert cpus is not None
        return cpus // 8 if cpus >= 16 else 2

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_argument(
            "-t",
            "--output_time",  # FIXME: --time
            help="(obsolete, time is always logged now) output the time it took for every step",
            action="store_true",
        )
        self.add_argument(
            "-d",
            "--debug",
            type=int,
            help="configure debug level; defaults to 0",
            default=1,
        )
        self.add_argument(
            "--log_file",
            help="configure where to write the log; defaults to stderr.",
            default="",
        )

    def add_pash_args(self):
        self.add_argument(
            "-w",
            "--width",
            type=int,
            default=self._get_width(),
            help="set data-parallelism factor",
        )
        self.add_argument(
            "--no_optimize",
            help="not apply transformations over the DFG",
            action="store_true",
        )
        self.add_argument(
            "--dry_run_compiler",
            help="not execute the compiled script, even if the compiler succeeded",
            action="store_true",
        )
        self.add_argument(
            "--assert_compiler_success",
            help="assert that the compiler succeeded with no general error occuring",
            action="store_true",
        )
        self.add_argument(
            "--assert_all_regions_parallelizable",
            help="assert that the compiler succeeded with all regions being parallelizable and no general error occuring (used to make tests more robust); more strict than --assert_compiler_success flag",
            action="store_true",
        )
        self.add_argument(
            "--avoid_pash_runtime_completion",
            help="avoid the pash_runtime execution completion (only relevant when --debug > 0)",
            action="store_true",
        )
        self.add_argument(
            "-p",
            "--output_optimized",  # FIXME: --print
            help="output the parallel shell script for inspection",
            action="store_true",
        )
        self.add_argument(
            "--graphviz",
            help="generates graphical representations of the dataflow graphs. The option argument corresponds to the format. PaSh stores them in a timestamped directory in the argument of --graphviz_dir",
            choices=["no", "dot", "svg", "pdf", "png"],
            default="no",
        )
        ## TODO: To discuss: Do we maybe want to have graphviz to always be included
        ##       in the temp directory (under a graphviz subdirectory) instead of in its own?
        ##   kk: I think that ideally we want a log-directory where we can put logs, graphviz,
        ##       and other observability and monitoring info (instead of putting them in the temp).
        self.add_argument(
            "--graphviz_dir",
            help="the directory in which to store graphical representations",
            default="/tmp",
        )
        self.add_argument(
            "--no_parallel_pipelines",
            help="Disable parallel running of independent pipelines",
            action="store_true",
            default=False,
        )
        self.add_argument(
            "--parallel_pipelines_limit",
            help="Maximum number of parallel independent pipelines",
            type=int,
            default=2,
        )
        self.add_argument(
            "--r_split_batch_size",
            type=int,
            help="configure the batch size of r_split (default: 1MB)",
            default=1000000,
        )
        self.add_argument(
            "--config_path",
            help="determines the config file path. By default it is 'PASH_TOP/compiler/config.yaml'.",
            default="",
        )
        self.add_argument(
            "--version",
            action="version",
            version="%(prog)s {version}".format(
                version="0.12.2"
            ),  # What does this version mean?
        )

        self.add_experimental_args()
        self.add_obsolete_args()

    def add_obsolete_args(self):
        self.add_argument(
            "--no_daemon",
            help="(obsolete) does nothing -- Run the compiler everytime we need a compilation instead of using the daemon",
            action="store_true",
            default=False,
        )
        self.add_argument(
            "--parallel_pipelines",
            help="(obsolete) Run multiple pipelines in parallel if they are safe to run. Now true by default. See --no_parallel_pipelines.",
            action="store_true",
            default=True,
        )
        self.add_argument(
            "--r_split",
            help="(obsolete) does nothing -- only here for old interfaces (not used anywhere in the code)",
            action="store_true",
        )
        self.add_argument(
            "--dgsh_tee",
            help="(obsolete) does nothing -- only here for old interfaces (not used anywhere in the code)",
            action="store_true",
        )

    def add_experimental_args(self):
        self.add_argument(
            "--no_eager",
            help="(experimental) disable eager nodes before merging nodes",
            action="store_true",
        )
        self.add_argument(
            "--profile_driven",
            help="(experimental) use profiling information when optimizing",
            action="store_true",
        )
        self.add_argument(
            "--speculative",
            help="(experimental) use the speculative execution preprocessing and runtime (NOTE: this has nothing to do with --speculation, which is actually misnamed, and should be named concurrent compilation/execution and is now obsolete)",
            action="store_true",
            default=False,
        )
        self.add_argument(
            "--termination",
            help="(experimental) determine the termination behavior of the DFG. Defaults to cleanup after the last process dies, but can drain all streams until depletion",
            choices=["clean_up_graph", "drain_stream"],
            default="clean_up_graph",
        )
        self.add_argument(
            "--daemon_communicates_through_unix_pipes",
            help="(experimental) the daemon communicates through unix pipes instead of sockets",
            action="store_true",
        )
        self.add_argument(
            "--distributed_exec",
            help="(experimental) execute the script in a distributed environment. Remote machines should be configured and ready",
            action="store_true",
            default=False,
        )



class CompilerParser(BaseParser):
    """
    Parser for the PaSh compiler in compiler/pash_compiler.py
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_pash_args()

        self.add_argument(
            "compiled_script_file",
            help="the file in which to output the compiled script",
        )
        self.add_argument(
            "input_ir",
            help="the file containing the dataflow graph to be optimized and executed",
        )
        self.add_argument(
            "--var_file",
            help="determines the path of a file containing all shell variables.",
            default=None,
        )



"""
PaSh Preprocessor Package

This package handles preprocessing of shell scripts for PaSh parallelization.
It identifies candidate dataflow regions and replaces them with calls to the
PaSh runtime for compilation and optimization.
"""

from preprocessor import preprocess

__all__ = ["preprocess"]



"""
DEPRECATED: This module exists for backward compatibility.
Use 'from compilation import ...' instead.

This module re-exports all symbols from compilation.py for backward compatibility.
It also maintains the main() function for CLI usage.
"""

import sys
import traceback

from sh_expand import env_vars_util

import config
from pash_graphviz import maybe_generate_graphviz
from util import log

from arg_parser import CompilerParser

# Re-export everything from compilation for backward compatibility
from compilation import (
    CompilerConfig,
    compile_ir,
    compile_optimize_output_script,
    load_df_region,
    compile_optimize_df_region,
    maybe_log_optimized_script,
    compile_candidate_df_region,
    optimize_irs,
    print_graph_statistics,
    choose_and_apply_parallelizing_transformations,
    choose_parallelizing_transformations,
    choose_parallelizing_transformation,
    apply_parallelizing_transformations,
    split_hdfs_cat_input,
    add_eager,
    add_eager_nodes,
)

# Import runtime_config from compilation (it's a module-level variable there)
import compilation
runtime_config = compilation.runtime_config


## We want to catch all exceptions here so that they are logged correctly
## and not just printed to the stderr.
def main():
    try:
        main_body()
    except Exception:
        log("Compiler failed, no need to worry, executing original script...")
        log(traceback.format_exc())
        sys.exit(1)


def main_body():
    ## Parse arguments
    args = parse_args()
    config.set_config_globals_from_pash_args(args)

    ## Load the configuration
    if not config.config:
        config.load_config(args.config_path)

    # Set runtime_config in compilation module
    compilation.runtime_config = config.config["distr_planner"]

    ## Read any shell variables files if present
    vars_dict = env_vars_util.read_vars_file(args.var_file, config.BASH_VERSION)
    config.set_vars_file(args.var_file, vars_dict)

    log("Input:", args.input_ir, "Compiled file:", args.compiled_script_file)

    ## Call the main procedure
    compiler_config = CompilerConfig(args.width)
    ast_or_ir = compile_optimize_output_script(
        args.input_ir, args.compiled_script_file, args, compiler_config
    )
    maybe_generate_graphviz(ast_or_ir, args)


def parse_args():
    parser = CompilerParser()
    args, _ = parser.parse_known_args()
    return args


if __name__ == "__main__":
    main()



"""
Setup script for PaSh with C binary compilation hooks.

This script handles the compilation of C runtime binaries during pip install.
The actual package configuration is in pyproject.toml.
"""

import os
import shutil
import subprocess
import sys
from pathlib import Path

from setuptools import setup
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop


def get_runtime_src_dir():
    """Get the runtime source directory."""
    return Path(__file__).parent / "src" / "pash" / "runtime"


def compile_runtime_binaries(build_lib=None):
    """Compile C runtime binaries using make.

    Args:
        build_lib: If provided, copy compiled binaries to this directory.
                   If None, compile in-place (for development mode).
    """
    runtime_src = get_runtime_src_dir()

    if not runtime_src.exists():
        print(f"Warning: Runtime source directory not found: {runtime_src}", file=sys.stderr)
        return

    # Check for required tools
    for tool in ["gcc", "make", "git"]:
        if shutil.which(tool) is None:
            raise RuntimeError(
                f"Required tool '{tool}' not found. "
                f"Please install build tools:\n"
                f"  Ubuntu/Debian: sudo apt install build-essential git\n"
                f"  Fedora/RHEL: sudo dnf install gcc make git\n"
                f"  macOS: xcode-select --install && brew install git"
            )

    # Run make in the runtime directory
    print(f"Compiling PaSh runtime binaries in {runtime_src}...")

    # Binary targets (including dgsh-tee which requires git to clone during build)
    targets = ["eager", "split", "r-merge", "r-wrap", "r-split", "r-unwrap", "set-diff", "dgsh-tee"]

    try:
        # Clean first to ensure fresh build
        subprocess.run(["make", "clean"], cwd=runtime_src, check=False, capture_output=True)

        # Compile
        result = subprocess.run(
            ["make"] + targets,
            cwd=runtime_src,
            check=True,
            capture_output=True,
            text=True,
        )
        if result.stdout:
            print(result.stdout)
    except subprocess.CalledProcessError as e:
        print(f"Make failed with output:\n{e.stdout}\n{e.stderr}", file=sys.stderr)
        raise RuntimeError(f"Failed to compile runtime binaries: {e}")

    # Copy binaries to target if different from source
    if build_lib:
        target_dir = Path(build_lib) / "pash" / "runtime"
        target_dir.mkdir(parents=True, exist_ok=True)

        # Binary names (note: make targets use hyphens, binaries use underscores)
        binaries = ["eager", "split", "r_split", "r_merge", "r_wrap", "r_unwrap", "set-diff", "dgsh-tee"]

        for binary in binaries:
            src_binary = runtime_src / binary
            if src_binary.exists():
                shutil.copy2(src_binary, target_dir / binary)
                print(f"  Copied {binary} to {target_dir}")
            else:
                print(f"  Warning: Binary {binary} not found at {src_binary}", file=sys.stderr)


class BuildPyWithRuntime(build_py):
    """Custom build_py that compiles C runtime binaries."""

    def run(self):
        # Run standard build_py first
        build_py.run(self)
        # Then compile runtime binaries
        compile_runtime_binaries(self.build_lib)


class DevelopWithRuntime(develop):
    """Custom develop that compiles C runtime binaries in place."""

    def run(self):
        # Compile runtime binaries in-place for development
        compile_runtime_binaries(None)
        # Run standard develop
        develop.run(self)


setup(
    cmdclass={
        "build_py": BuildPyWithRuntime,
        "develop": DevelopWithRuntime,
    },
    # Don't mark as having ext_modules - we compile from source at install time
    # When users install from the sdist, build_py will compile the C binaries
)



import config
import os
import uuid

from util import string_to_argument, make_kv
from ir_defs.resource import (
    Resource,
    FileDescriptorResource,
    FileResource,
    TemporaryFileResource,
    EphemeralResource,
    RemoteFileResource,
    HDFSFileResource,
    DFSSplitResource,
)


## Note: The NULL ident is considered to be the default unknown file id
##
## TODO: WARNING: We have to make sure that a resource in our IR can
## be uniquely determined given a relative or absolute path. Actually,
## we need to make sure that expanding any variable/string in our IR,
## will always return the same result.
##
## WARNING: At the moment it is not clear what resources are saved in
## the Find(self) and in self. This might create problems down the
## road.
##
## TODO: When doing union, I have to really make both file ids point
## to the same file.
class FileId:
    def __init__(self, ident, prefix="", resource=None):
        self.ident = ident
        ## TODO: Add this as part of the resource. Ephemeral resources should
        ##       have a prefix. Normal resources have a file system id. Descriptor
        ##       resources have a number.
        self.prefix = prefix
        ## TODO: Remove all union_find
        ## Initialize the parent
        self.resource = resource

    def __repr__(self):
        if isinstance(self.resource, EphemeralResource):
            output = self.get_fifo_suffix()
        else:
            output = "fid:{}:{}".format(self.ident, self.resource)
        return output

    def serialize(self):
        if isinstance(self.resource, TemporaryFileResource):
            output = self.get_temporary_file_suffix()
        elif isinstance(self.resource, EphemeralResource):
            output = self.get_fifo_suffix()
        else:
            output = "{}".format(self.resource)
        return output

    def get_temporary_file_suffix(self):
        tempfile_name = "{}{}".format(self.prefix, self.ident)
        return tempfile_name

    def get_fifo_suffix(self):
        fifo_name = "{}#fifo{}".format(self.prefix, self.ident)
        return fifo_name

    ## Serialize as an option for the JSON serialization when sent to
    ## the backend. This is needed as options can either be files or
    ## arguments, and in each case there needs to be a different
    ## serialization procedure.
    def opt_serialize(self):
        return '"{}"'.format(self.serialize())

    ## Returns a shell AST from this file identifier.
    ## TODO: Once the python libdash bindings are done we could use those instead.
    ##
    ## If stdin_dash is True, then we can turn stdin to `-` since the
    ##   context in which we are using it is a command for which `-` means stdin.
    def to_ast(self, stdin_dash=False):
        ## TODO: This here is supposed to identify fifos, but real fifos have a resource
        ##       but are fifos. Therefore eventually we want to have this check correctly
        ##       check if a file id refers to a pipe
        ##
        ## TODO: I am not sure about the FileDescriptor resource
        if isinstance(self.resource, TemporaryFileResource):
            suffix = self.get_temporary_file_suffix()
            string = os.path.join(config.PASH_TMP_PREFIX, suffix)
            argument = string_to_argument(string)
        elif isinstance(self.resource, EphemeralResource):
            suffix = self.get_fifo_suffix()
            string = os.path.join(config.PASH_TMP_PREFIX, suffix)
            ## Quote the argument
            argument = [make_kv("Q", string_to_argument(string))]
        elif isinstance(self.resource, FileDescriptorResource):
            if self.resource.is_stdin() and stdin_dash:
                argument = string_to_argument("-")
            else:
                raise NotImplementedError()
        else:
            ## The resource holds an Argument, and we need to get its argument list
            argument = self.resource.uri.to_ast()

        return argument

    def set_resource(self, resource):
        ## The file resource cannot be reset. A pointer can never point to
        ## more than one file resource. However, we can change an ephemeral
        ## resource or a file_descriptor resource.
        assert not self.has_file_resource()
        self.resource = resource

    def get_resource(self):
        return self.resource

    ## Remove this
    def has_resource(self):
        return not self.resource is None

    def has_file_resource(self):
        return isinstance(self.resource, FileResource)

    def has_file_descriptor_resource(self):
        return isinstance(self.resource, FileDescriptorResource)

    def has_remote_file_resource(self):
        return isinstance(self.resource, RemoteFileResource)

    def is_ephemeral(self):
        return isinstance(self.resource, EphemeralResource)

    def make_temporary_file(self):
        self.resource = TemporaryFileResource()

    ## Removes a resource from an FID, making it ephemeral
    def make_ephemeral(self):
        self.resource = EphemeralResource()

    def toFileName(self, prefix):
        output = "{}_file{}".format(prefix, self.ident)
        return output

    def isNull(self):
        return self.ident == "NULL"

    def get_ident(self):
        return self.ident

    def is_available_on(self, host):
        if self.is_ephemeral():
            return True
        elif self.has_remote_file_resource():
            return self.resource.is_available_on(host)
        else:
            # Currently any other resource types should
            # be part of the main shell graph.
            return False



#!/usr/bin/env python3
"""
PaSh CLI Entry Point

This module provides the `pash` command-line interface as a wrapper
around pa.sh. When installed via pip, users can run `pash` instead
of `./pa.sh`.

The actual execution is delegated to pa.sh which handles:
1. Environment setup (shell functions, FIFOs, etc.)
2. Starting the compilation server
3. Running the preprocessor
4. Executing the preprocessed script
5. Cleanup
"""

import os
import subprocess
import sys
from pathlib import Path

from . import __version__


def get_pash_top() -> Path:
    """Get the PASH_TOP directory (package installation path)."""
    return Path(__file__).parent


def get_pa_sh_path() -> Path:
    """Get the path to pa.sh.

    pa.sh is located in the same directory as this file (src/pash/).
    """
    pash_top = get_pash_top()
    pa_sh = pash_top / "pa.sh"

    if pa_sh.exists():
        return pa_sh

    # Fallback: check PASH_TOP environment variable
    pash_top_env = os.environ.get("PASH_TOP")
    if pash_top_env:
        env_pa_sh = Path(pash_top_env) / "pa.sh"
        if env_pa_sh.exists():
            return env_pa_sh

    raise FileNotFoundError(
        "Could not find pa.sh. Make sure PaSh is properly installed."
    )


def main():
    """Main entry point for PaSh CLI.

    This delegates to pa.sh for actual execution, passing through all arguments.
    """
    # Handle --version specially
    if len(sys.argv) == 2 and sys.argv[1] in ("--version", "-V"):
        print(f"pash {__version__}")
        sys.exit(0)

    try:
        pa_sh_path = get_pa_sh_path()
    except FileNotFoundError as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)

    # Set PASH_TOP to point to the src/pash directory
    env = os.environ.copy()
    env["PASH_TOP"] = str(get_pash_top())

    # Pass all arguments to pa.sh
    args = ["bash", str(pa_sh_path)] + sys.argv[1:]

    try:
        result = subprocess.run(args, env=env)
        sys.exit(result.returncode)
    except KeyboardInterrupt:
        sys.exit(130)
    except Exception as e:
        print(f"Error running PaSh: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import os
from datetime import datetime

from util import (
    print_time_delta,
    string_to_argument,
    standard_var_ast,
    make_quoted_variable,
    make_command,
    make_semi_sequence,
    make_defun,
    make_subshell,
    redir_append_stderr_to_string_file,
)
from shasta.json_to_ast import to_ast_node
from parse import from_ast_objects_to_shell
import config

RM_PASH_FIFOS_NAME = "rm_pash_fifos"
MKFIFO_PASH_FIFOS_NAME = "mkfifo_pash_fifos"


def to_shell(ir, args):
    backend_start_time = datetime.now()

    ## First call an IR to AST compilation pass
    output_asts = ir2ast(ir, args)

    ## Then just call the parser.
    output_script = from_ast_objects_to_shell(output_asts)

    backend_end_time = datetime.now()
    print_time_delta("Backend", backend_start_time, backend_end_time)

    return output_script


def ir2ast(ir, args):
    clean_up_graph = False
    drain_streams = False
    if args.termination == "clean_up_graph":
        clean_up_graph = True
    elif args.termination == "drain_stream":
        drain_streams = True

    ## NOTE: We first need to make the main body because it might create additional ephemeral fids.

    ## TODO: If we have just a single node maybe we should just instantiate that without anything else.
    ## This is implemented below, but there is no need to turn it on now.
    ##
    ## If we only have a single node, then we don't need to make prologues, epilogues, etc
    # if len(ir.nodes) == 1:
    #     nodes = list(ir.nodes.values())
    #     assert(len(nodes) == 1)
    #     node = nodes[0]
    #     body = [node.to_ast(ir.edges, drain_streams)]
    #     return body

    ## Make the main body
    body = ir.to_ast(drain_streams)

    all_fids = ir.all_fids()

    # log("All fids:", all_fids)
    ## Find all the ephemeral fids and turn them to ASTs
    ephemeral_fids = [fid for fid in all_fids if fid.is_ephemeral()]

    # log("Ephemeral fids:", ephemeral_fids)

    ## Call the prologue that creates fifos for all ephemeral fids
    prologue = make_ir_prologue(ephemeral_fids)

    ## Call the epilogue that removes all ephemeral fids
    epilogue = make_ir_epilogue(ephemeral_fids, clean_up_graph, args.log_file)

    final_asts = prologue + body + epilogue

    return final_asts


def make_rms_f_prologue_epilogue(ephemeral_fids):
    asts = []
    ## Create an `rm -f` for each ephemeral fid
    for eph_fid in ephemeral_fids:
        args = [eph_fid.to_ast()]
        command = make_rm_f_ast(args)
        asts.append(command)
    return asts


def make_ir_prologue(ephemeral_fids) -> "list[AstNode]":
    asts = []
    ## Create an `rm -f` for each ephemeral fid
    rm_asts = make_rms_f_prologue_epilogue(ephemeral_fids)
    defun_rm_pash_fifos = make_defun(RM_PASH_FIFOS_NAME, make_semi_sequence(rm_asts))
    asts.append(defun_rm_pash_fifos)

    ## Create a `mkfifo` for each ephemeral fid
    mkfifo_asts = []
    for eph_fid in ephemeral_fids:
        args = [eph_fid.to_ast()]
        command = make_mkfifo_ast(args)
        mkfifo_asts.append(command)

    defun_mkfifos = make_defun(MKFIFO_PASH_FIFOS_NAME, make_semi_sequence(mkfifo_asts))
    asts.append(defun_mkfifos)

    call_rm_pash_fifos = make_command([string_to_argument(RM_PASH_FIFOS_NAME)])
    asts.append(call_rm_pash_fifos)

    call_mkfifos = make_command([string_to_argument(MKFIFO_PASH_FIFOS_NAME)])
    asts.append(call_mkfifos)

    class_asts = [to_ast_node(ast) for ast in asts]
    return class_asts


def make_ir_epilogue(ephemeral_fids, clean_up_graph, log_file) -> "list[AstNode]":
    asts = []
    if clean_up_graph:
        ## TODO: Wait for all output nodes not just one
        pids = [[standard_var_ast("!")]]
        clean_up_path_script = os.path.join(
            config.PASH_TOP, config.config["runtime"]["clean_up_graph_binary"]
        )
        com_args = [
            string_to_argument("source"),
            string_to_argument(clean_up_path_script),
        ] + pids
        if log_file == "":
            com = make_command(com_args)
        else:
            redirection = redir_append_stderr_to_string_file(log_file)
            com = make_command(com_args, redirections=[redirection])
        asts.append(com)
    else:
        ## Otherwise we just wait for all processes to die.
        wait_com = make_command([string_to_argument("wait")])
        exit_status = make_command([string_to_argument("internal_exec_status=$?")])
        asts.extend([wait_com, exit_status])

    ## Create an `rm -f` for each ephemeral fid
    call_rm_pash_funs = make_command([string_to_argument(RM_PASH_FIFOS_NAME)])
    asts.append(call_rm_pash_funs)

    ## Make the following command:
    #    (exit $internal_exec_status)
    exit_ec_ast = make_exit_ec_ast()
    asts.append(exit_ec_ast)

    class_asts = [to_ast_node(ast) for ast in asts]
    return class_asts


def make_exit_ec_ast():
    command = make_command(
        [string_to_argument("exit"), [make_quoted_variable("internal_exec_status")]]
    )
    ast = make_subshell(command)
    return ast


def make_rm_f_ast(arguments):
    all_args = [string_to_argument("rm"), string_to_argument("-f")] + arguments
    return make_command(all_args)


def make_mkfifo_ast(arguments):
    all_args = [string_to_argument("mkfifo")] + arguments
    return make_command(all_args)



from util import log
from ir_defs.resource import FileResource, Resource, FileDescriptorResource
from pash_annotations.datatypes.BasicDatatypesWithIO import (
    FileNameWithIOInfo,
    StdDescriptorWithIOInfo,
)


def resource_from_file_descriptor(file_descriptor) -> Resource:
    if isinstance(file_descriptor, FileNameWithIOInfo):
        arg = file_descriptor.get_name()
        log(f"filedes name: {file_descriptor.get_name()}")
        log(f"filedes name type: {type(file_descriptor.get_name())}")
        log(f"arg: {arg}")
        return FileResource(file_descriptor.get_name())
    elif isinstance(file_descriptor, StdDescriptorWithIOInfo):
        resource = ("fd", file_descriptor.get_type().value)
        return FileDescriptorResource(resource)
    else:
        assert False
        # unreachable



"""
PaSh Runtime Module

This module provides utilities for finding compiled runtime binaries.
The binaries are compiled from C source during pip install.

Available binaries:
- eager: Eager evaluation node
- split: Data splitting
- r_split: Record-aware splitting
- r_merge: Record-aware merging
- r_wrap: Record wrapping
- r_unwrap: Record unwrapping
- set-diff: Set difference
"""

from pathlib import Path


def get_runtime_dir() -> Path:
    """Get the runtime directory containing binaries."""
    return Path(__file__).parent


def get_binary_path(name: str) -> Path:
    """Get the path to a runtime binary.

    Args:
        name: Name of the binary (e.g., 'eager', 'split')

    Returns:
        Path to the binary

    Raises:
        FileNotFoundError: If the binary doesn't exist
    """
    binary = get_runtime_dir() / name
    if not binary.exists():
        raise FileNotFoundError(
            f"Runtime binary '{name}' not found at {binary}. "
            f"You may need to reinstall pash with 'pip install --force-reinstall pash'"
        )
    return binary


# Convenience functions for each binary
def get_eager_binary() -> Path:
    """Get path to the eager binary."""
    return get_binary_path("eager")


def get_split_binary() -> Path:
    """Get path to the split binary."""
    return get_binary_path("split")


def get_r_split_binary() -> Path:
    """Get path to the r_split binary."""
    return get_binary_path("r_split")


def get_r_merge_binary() -> Path:
    """Get path to the r_merge binary."""
    return get_binary_path("r_merge")


def get_r_wrap_binary() -> Path:
    """Get path to the r_wrap binary."""
    return get_binary_path("r_wrap")


def get_r_unwrap_binary() -> Path:
    """Get path to the r_unwrap binary."""
    return get_binary_path("r_unwrap")


def get_set_diff_binary() -> Path:
    """Get path to the set-diff binary."""
    return get_binary_path("set-diff")



import socket, pickle
from threading import Thread
from socket_utils import encode_request, decode_request
import subprocess
import json
import sys
import os
import asyncio
import shlex
import time
import uuid
import argparse

PASH_TOP = os.environ["PASH_TOP"]
sys.path.append(os.path.join(PASH_TOP, "compiler"))

import config
from arg_parser import BaseParser
from util import log
import pash_compiler
from dspash.socket_utils import send_msg, recv_msg
from dspash.ir_helper import save_configs, to_shell_file
from dspash.utils import create_filename, write_file

# from ... import config
HOST = socket.gethostbyname(socket.gethostname())
PORT = 65432  # Port to listen on (non-privileged ports are > 1023)


def err_print(*args):
    print(*args, file=sys.stderr)


def send_success(conn, body, msg=""):
    request = {"status": "OK", "body": body, "msg": msg}
    send_msg(conn, encode_request(request))


def parse_exec_request(request):
    return request["cmd"]


def parse_exec_graph(request):
    return request["graph"], request["shell_variables"], request["functions"]


def exec_graph(graph, shell_vars, functions):
    config.config["shell_variables"] = shell_vars
    script_path = to_shell_file(graph, config.pash_args)

    e = os.environ.copy()
    e["PASH_TOP"] = PASH_TOP

    # store functions
    functions_file = create_filename(dir=config.PASH_TMP_PREFIX, prefix="pashFuncs")
    write_file(functions_file, functions)
    cmd = f"source {functions_file}; source {script_path}"
    rc = subprocess.Popen(cmd, env=e, executable="/bin/bash", shell=True)
    return rc


class Worker:
    def __init__(self, port=None):
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if port == None:
            # pick a random port
            self.s.bind((HOST, 0))
        else:
            self.s.bind((HOST, port))
        print(f"Worker running on {HOST}:{self.s.getsockname()[1]}")

    def run(self):
        connections = []
        with self.s:
            self.s.listen()
            while True:
                conn, addr = self.s.accept()
                print(f"got new connection")
                t = Thread(target=manage_connection, args=[conn, addr])
                t.start()
                connections.append(t)
        for t in connections:
            t.join()


def manage_connection(conn, addr):
    rcs = []
    with conn:
        print("Connected by", addr)
        dfs_configs_paths = {}
        while True:
            data = recv_msg(conn)
            if not data:
                break

            print("got new request")
            request = decode_request(data)
            if request["type"] == "Exec-Graph":
                graph, shell_vars, functions = parse_exec_graph(request)
                save_configs(graph, dfs_configs_paths)
                exec_graph(graph, shell_vars, functions)
                body = {}
            else:
                print(f"Unsupported request {request}")
            send_success(conn, body)
    print("connection ended")
    for rc in rcs:
        rc.wait()


def parse_args():
    parser = BaseParser(description="PaSh distributed worker.")
    parser.add_pash_args()
    parser.add_argument("--port", type=int, help="port to use", default=65432)
    args = parser.parse_args()
    config.set_config_globals_from_pash_args(args)
    if not config.config:
        config.load_config(args.config_path)
    return args


def init():
    args = parse_args()
    config.LOGGING_PREFIX = f"Worker {config.pash_args.port}: "
    ## KK: 2023-02-21 Commenting this out, we need to figure out if the new annotations work with the distribution package
    # config.annotations = load_annotation_files(
    #     config.config['distr_planner']['annotations_dir'])
    pash_compiler.runtime_config = config.config["distr_planner"]
    pash_compiler.termination = ""


def main():
    init()
    worker = Worker(config.pash_args.port)
    worker.run()


if __name__ == "__main__":
    main()
